# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, List, Dict

from chipscopy import dm
from chipscopy.client import core
from chipscopy.client.core import CoreClient
from chipscopy.dm.request import DoneCallback
from chipscopy.utils import listify
from chipscopy.utils.logger import log

DOMAIN_NAME = "client_ibert"


class IBERTCoreClient(CoreClient):
    @staticmethod
    def is_compatible(node: dm.Node) -> bool:
        return_val = node.type and node.type == "ibert"
        return return_val

    def post_init(self):
        super(IBERTCoreClient, self).post_init()

    def get_child_with_name(self, name: str, done: DoneCallback):
        """
        Iterate through the children of this IBERT and return Node with matching name.
        """
        for child_ctx in self.children:
            child_node = self.manager.get_node(child_ctx)
            if name == child_node.props["display_name"]:
                done(result=child_node)
                return

        done()

    def get_service_proxy(self):
        return self.manager.channel.getRemoteService("IBERT")

    def initialize(self, done: DoneCallback = None):
        service, done_cb = self.make_done(done)
        log[DOMAIN_NAME].info(f"Initializing IBERT service")
        token = service.initialize(self.ctx, done_cb)
        return self.add_pending(token)

    def setup(self, done: DoneCallback = None):
        service, done_cb = self.make_done(done)
        options = {"node_id": self.ctx}
        token = service.setup(options, done_cb)
        return self.add_pending(token)

    def read(
        self,
        start_address: int,
        endpoint_display_name: str,
        size: int = 1,
        *,
        done: DoneCallback = None,
    ):
        """
        Command to do single or burst read.

        Args:
            start_address (int): The address to read from.

            endpoint_display_name (str): Display name for the endpoint the register is located in.

            size (int): **(Optional)** The number of words to read. Default = 1.

            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        Returns:
            str:
                Format - `"{Address as hex} : {Value in hex} : {Value as bin}"`

        Usage
            ::

                #
                >>> ibert.read(0x3000, "Quad_0")
                0x3000 : 0x0000_1234 : 0b0000_0000_0000_0000_0001_0010_0011_0100
                >>> ibert.read(0x3000, "Quad_0", 5)
                0x3000 : 0x0000_1234 : 0b0000_0000_0000_0000_0001_0010_0011_0100
                0x3004 : 0x0000_0000 : 0b0000_0000_0000_0000_0000_0000_0000_0000
                0x3008 : 0x0870_0000 : 0b0000_1000_0111_0000_0000_0000_0000_0000
                0x300C : 0x0000_0000 : 0b0000_0000_0000_0000_0000_0000_0000_0000
                0x3010 : 0x0000_0000 : 0b0000_0000_0000_0000_0000_0000_0000_0000

        """
        service, done_cb = self.make_done(done)
        options = {
            "node_id": self.ctx,
            "endpoint_name": endpoint_display_name,
            "start": start_address,
            "size": size,
        }
        token = service.read(options, done_cb)
        return self.add_pending(token)

    def write(
        self,
        start_address: int,
        data: int,
        endpoint_display_name: str,
        *,
        done: DoneCallback = None,
    ):
        """
        Command to write a value to an address.

        Args:
            start_address (int): Address to write to.

            data (int): Value to write.

            endpoint_display_name (str): Display name for the endpoint the register is located in.

            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        Returns:
            None

        Usage
            ::

                #
                >>> ibert.write(0x1234, 0xABCD_EF12, "Quad_0")
                None

        """
        service, done_cb = self.make_done(done)
        options = {
            "node_id": self.ctx,
            "endpoint_name": endpoint_display_name,
            "start": start_address,
            "data": data,
        }
        token = service.write(options, done_cb)
        return self.add_pending(token)

    def get_layout(self, *, done: DoneCallback = None):
        """
        Get the hierarchy of objects in the IBERT node and the properties associated at each level.

        Args:
            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        Returns:
            dict: Nested dictionaries with 2 keys at every level - "Properties" and "Children".

        """
        service, done_cb = self.make_done(done)
        options = {"node_id": self.ctx}
        token = service.get_layout(options, done_cb)
        return self.add_pending(token)

    def get_property(
        self,
        property_names: Union[str, List[str]],
        endpoint_name: str,
        *,
        done: DoneCallback = None,
    ):
        """
        Get value of the properties from the core object.

        Args:
            property_names (str or list[str]): A string with the name of the property OR
                a list of property name string(s).

            endpoint_name (str): Display name of the endpoint to which the
                property(properties) belong.

            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        Returns:
            dict(str, str) or str: String or dictionary depending on how the command is invoked.
            See below for examples.

        Usage
            ::

                # String as first argument
                >>> ibert.get_property("<Property name>", "Quad_0")
                "<Property value">

                # List of length 1 as first argument
                >>> ibert.get_property(["<Property name>"], "Quad_0")
                "<Property value">

                # List of length 3 as first argument
                >>> ibert.get_property(
                >>>     [
                >>>         "<Property 1 name>",
                >>>         "<Property 1 name>",
                >>>         "<Property n name>",
                >>>     ],
                >>>     "Quad_0"
                >>> )
                {
                    "<Property 1 name>": "<Property value>",
                    "<Property 2 name>": "<Property value>",
                    "<Property n name>": "<Property value>",
                }

        """
        service, done_cb = self.make_done(done)
        if isinstance(property_names, str):
            property_names = [property_names]
        options = {
            "node_id": self.ctx,
            "Property Names": property_names,
            "Endpoint Display Name": endpoint_name,
        }
        token = service.get_property(options, done_cb)
        return self.add_pending(token)

    def get_property_group(self, groups: Union[str, List[str]], *, done: DoneCallback = None):
        """
        Get value of the properties in property group(s) from the core object.

        Args:
            groups (list[str] or str):  A string with the name of the property group OR
                a list of property group string(s).

            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        Returns:
            dict: Dictionary with depth of 1 or 2, depending on how the command is invoked.
            See below for examples

        Usage
            ::

                # String as argument
                >>> ibert.get_property_group("<Property group name>")
                {
                    "<Property 1 name>": "<Property value>",
                    "<Property 2 name>": "<Property value>",
                    "<Property n name>": "<Property value>",
                }

                # List of length 1 as argument
                >>> ibert.get_property_group(["<Property group name>"])
                {
                    "<Property 1 name>": "<Property value>",
                    "<Property 2 name>": "<Property value>",
                    "<Property n name>": "<Property value>",
                }

                # List of length 3 as argument
                >>> ibert.get_property_group(
                >>>    [
                >>>         "<Property group 1 name>",
                >>>         "<Property group 2 name>",
                >>>         "<Property group n name>"
                >>>    ]
                >>> )
                {
                    "<Property group 1 name>":  {
                        "<Property 1 name>": "<Property value>",
                        "<Property 2 name>": "<Property value>",
                        "<Property n name>": "<Property value>",
                    },
                    "<Property group 2 name>": {
                        "<Property 1 name>": "<Property value>",
                        "<Property 2 name>": "<Property value>",
                        "<Property n name>": "<Property value>",
                    },
                    "<Property group n name>": {
                        "<Property 1 name>": "<Property value>",
                        "<Property 2 name>": "<Property value>",
                        "<Property n name>": "<Property value>",
                    }
                }

        """
        service, done_cb = self.make_done(done)
        if isinstance(groups, str):
            groups = [groups]
        options = {"node_id": self.ctx, "Groups": groups}
        token = service.get_property_group(options, done_cb)
        return self.add_pending(token)

    def set_property(
        self,
        property_dict: Dict[str, Union[int, str]],
        endpoint_name: str,
        *,
        done: DoneCallback = None,
    ):
        """
        Set a new value for the property in the core object.

        Args:
            property_dict (dict): A dictionary with key, value pair as the property name and
                new value for property.

            endpoint_name (str): Display name of the endpoint to which the
                property(properties) belong.

            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        Returns:
            None

        Usage
            ::

                #
                >>> ibert.set_property(
                >>>     {
                >>>         "<Property 1>": "<New value>",
                >>>         "<Property 2>": "<New value>",
                >>>         "<Property n>": "<New value>",
                >>>     },
                >>>     "Quad_0"
                >>> )
                None

        """
        service, done_cb = self.make_done(done)
        options = {
            "node_id": self.ctx,
            "Property Dict": property_dict,
            "Endpoint Display Name": endpoint_name,
        }
        token = service.set_property(options, done_cb)
        return self.add_pending(token)

    def refresh_property(
        self,
        property_names: Union[str, List[str]],
        endpoint_name: str,
        *,
        done: DoneCallback = None,
    ):
        """
        Update value of properties from the core.

        Args:
            property_names (str or list(str)): A string with the name of the property OR a list
                of property name string(s).

            endpoint_name (str): Display name of the endpoint to which the
                property(properties) belong.

            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        Returns:
            dict(str, str) or str: String or dictionary depending on how the command is invoked.
            See below for examples.

        Usage
            ::

                # String as first argument
                >>> ibert.refresh_property("<Property name>", "Quad_0")
                "<Refreshed value>"

                # List of length 1 as first argument
                >>> ibert.refresh_property(["<Property name>"], "Quad_0")
                "<Refreshed value>"

                # List of length 3 as first argument
                >>> ibert.refresh_property(
                >>>     [
                >>>         "<Property 1 name>",
                >>>         "<Property 2 name>",
                >>>         "<Property n name>"
                >>>     ],
                >>>     "Quad_0"
                >>> )
                {
                    "<Property 1 name>": "<Refreshed value>",
                    "<Property 2 name>": "<Refreshed value>",
                    .
                    .
                    .
                    "<Property n name>": "<Refreshed value>",
                }

        """
        service, done_cb = self.make_done(done)
        if isinstance(property_names, str):
            property_names = [property_names]
        options = {
            "node_id": self.ctx,
            "Property Names": property_names,
            "Endpoint Display Name": endpoint_name,
        }
        token = service.refresh_property(options, done_cb)
        return self.add_pending(token)

    def refresh_property_group(self, groups: Union[str, List[str]], *, done: DoneCallback = None):
        """
        Update the value of the properties in the property group(s) by reading from the core.

        Args:
            groups (str or list[str]): The group name to refresh as a string OR a list of group
                name as string(s).

            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        Returns:
            dict - Dictionary with a depth of 1 or 2, depending on how the command is invoked.
            See below for examples.

        Usage
            ::

                # String as argument
                >>> ibert.refresh_property_group("<Property group name>")
                {
                    "<Property 1 name>": "<Refreshed value>",
                    "<Property 2 name>": "<Refreshed value>",
                    "<Property n name>": "<Refreshed value>",
                }

                # List of length 1 as argument
                >>> ibert.refresh_property_group(["<Property group name>"])
                {
                    "<Property 1 name>": "<Refreshed value>",
                    "<Property 2 name>": "<Refreshed value>",
                    "<Property n name>": "<Refreshed value>",
                }

                # List of length 3 as argument
                >>> ibert.refresh_property_group(
                >>>    [
                >>>         "<Property group 1 name>",
                >>>         "<Property group 2 name>",
                >>>         "<Property group n name>"
                >>>    ]
                >>> )
                {
                    "<Property group 1 name>": {
                        "<Property 1 name>": "<Refreshed value>",
                        "<Property 2 name>": "<Refreshed value>",
                        "<Property n name>": "<Refreshed value>",
                    },
                    "<Property group 2 name>": {
                        "<Property 1 name>": "<Refreshed value>",
                        "<Property 2 name>": "<Refreshed value>",
                        "<Property n name>": "<Refreshed value>",
                    },
                    "<Property group n name>": {
                        "<Property 1 name>": "<Refreshed value>",
                        "<Property 2 name>": "<Refreshed value>",
                        "<Property n name>": "<Refreshed value>",
                    }
                }

        """
        service, done_cb = self.make_done(done)
        if isinstance(groups, str):
            groups = [groups]
        options = {"node_id": self.ctx, "Groups": groups}
        token = service.refresh_property_group(options, done_cb)
        return self.add_pending(token)

    def commit_property(
        self,
        property_names: Union[str, List[str]],
        endpoint_name: str,
        *,
        done: DoneCallback = None,
    ):
        """
        Commit the current value of the properties in the object to the core.

        Args:
            property_names (str or list[str]): The property name to commit as a string OR a list
                of property name string(s).

            endpoint_name (str): Display name of the endpoint to which the
                property(properties) belong.

            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        Returns:
            None

        Usage
            ::

                #
                >>> ibert.commit_property("<Property name>", "Quad_0")
                None

                >>> ibert.commit_property(
                >>>     [
                >>>         "<Property 1 name">,
                >>>         "<Property 2 name">,
                >>>         "<Property n name">,
                >>>     ],
                >>>     "Quad_0"
                >>> )
                None

        """
        service, done_cb = self.make_done(done)
        if isinstance(property_names, str):
            property_names = [property_names]
        options = {
            "node_id": self.ctx,
            "Property Names": property_names,
            "Endpoint Display Name": endpoint_name,
        }
        token = service.commit_property(options, done_cb)
        return self.add_pending(token)

    def list_property_groups(self, *, done: DoneCallback = None):
        """
        Lists the available property groups.

        Args:
            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        Returns:
            list(str): A list of group names

        Usage
            ::

                #
                >>> ibert.list_property_groups
                ["<Property group 1 name>", "<Property group 2 name>", "<Property group n name>"]

        """
        service, done_cb = self.make_done(done)
        options = {"node_id": self.ctx}
        token = service.list_property_groups(options, done_cb)
        return self.add_pending(token)

    def add_to_property_watchlist(
        self,
        property_names: Union[str, List[str]],
        endpoint_name: str,
        *,
        done: DoneCallback = None,
    ):
        service, done_cb = self.make_done(done)
        property_names = listify(property_names)
        options = {
            "node_id": self.ctx,
            "Property Names": property_names,
            "Endpoint Display Name": endpoint_name,
        }
        token = service.add_to_property_watchlist(options, done_cb)
        return self.add_pending(token)

    def remove_from_property_watchlist(
        self,
        property_names: Union[str, List[str]],
        endpoint_name: str,
        *,
        done: DoneCallback = None,
    ):
        service, done_cb = self.make_done(done)
        property_names = listify(property_names)
        options = {
            "node_id": self.ctx,
            "Property Names": property_names,
            "Endpoint Display Name": endpoint_name,
        }
        token = service.remove_from_property_watchlist(options, done_cb)
        return self.add_pending(token)

    def report_property(
        self,
        property_names: Union[str, List[str]],
        endpoint_name: str,
        *,
        done: DoneCallback = None,
    ):
        """
        Get detailed report about a property/properties.

        Args:
            property_names (str or list(str)): The name of the property to get report for a list
                of property name string(s).

            endpoint_name (str): Display name of the endpoint to which the
                property(properties) belong.

            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        Returns:
            dict: The report for a property may vary based on the type of properties

        Usage
            ::

                # Report property on a regular property
                >>> ibert.report_property("Property A", "Quad_0")
                {
                    "Property A": {
                        "Name": "Property A",
                        "Description: "This is a regular property, that can have any value",
                        "Permission": [
                            "GET",
                            "SET",
                            "COMMIT",
                            "REFRESH"
                        ],
                        "Default value": "0",
                        "Groups": [
                            "Group 1",
                            "Group 2"
                        ],
                        "Current value": "23",
                        "Value type": str
                    }
                }

                # Property B is an enumerated property
                # In this case, an additional key "Valid values" is included in the report
                >>> ibert.report_property("Property B", "Quad_0")
                {
                    "Property B":  {
                        "Name": "Property B",
                        "Description: "This is an enumerated property",
                        "Permission": [
                            "GET",
                            "SET",
                            "COMMIT",
                            "REFRESH"
                        ],
                        "Default value": "0",
                        "Groups": [
                            "Group 1",
                            "Group 2"
                        ],
                        "Current value": "23",
                        "Valid values": [
                            "0",
                            "10",
                            "14",
                            "23",
                            "50"
                        ],
                        "Value type": str
                    }
                }

        """
        service, done_cb = self.make_done(done)
        if isinstance(property_names, str):
            property_names = [property_names]
        options = {
            "node_id": self.ctx,
            "Property Names": property_names,
            "Endpoint Display Name": endpoint_name,
        }
        token = service.report_property(options, done_cb)
        return self.add_pending(token)

    def start_eye_scan(
        self,
        rx_name: str,
        scan_parameters=None,
        *,
        done: DoneCallback = None,
    ):
        service, done_cb = self.make_done(done)

        if scan_parameters is None:
            scan_parameters = dict()

        options = {"node_id": self.ctx, "RX Name": rx_name, "Scan Parameters": scan_parameters}

        token = service.start_eye_scan(options, done_cb)
        return self.add_pending(token)

    def terminate_eye_scan(self, rx_name: str, *, done: DoneCallback = None):
        service, done_cb = self.make_done(done)

        options = {"node_id": self.ctx, "RX Name": rx_name}

        token = service.terminate_eye_scan(options, done_cb)
        return self.add_pending(token)

    def get_eye_scan_parameters(self, rx_name: str = None, *, done: DoneCallback = None):
        service, done_cb = self.make_done(done)

        if rx_name is None:
            rx_name = list()

        if isinstance(rx_name, str):
            rx_name = [rx_name]

        options = {"node_id": self.ctx, "RX Name": rx_name}
        token = service.get_eye_scan_parameters(options, done_cb)
        return self.add_pending(token)

    def start_yk_scan(self, rx_name: str, *, done: DoneCallback = None):
        service, done_cb = self.make_done(done)

        options = {"node_id": self.ctx, "RX Name": rx_name}

        token = service.start_yk_scan(options, done_cb)
        return self.add_pending(token)

    def terminate_yk_scan(self, rx_name: str, *, done: DoneCallback = None):
        service, done_cb = self.make_done(done)

        options = {"node_id": self.ctx, "RX Name": rx_name}

        token = service.terminate_yk_scan(options, done_cb)
        return self.add_pending(token)



# %% [markdown]
# <link rel="preconnect" href="https://fonts.gstatic.com">
# <link href="https://fonts.googleapis.com/css2?family=Fira+Code&display=swap" rel="stylesheet">
#
# ### License
#
# <p style="font-family: 'Fira Code', monospace; font-size: 1.2rem">
# Copyright 2021 Xilinx, Inc.<br><br>
# Licensed under the Apache License, Version 2.0 (the "License");<br>
# you may not use this file except in compliance with the License.<br><br>
# You may obtain a copy of the License at <a href="http://www.apache.org/licenses/LICENSE-2.0"?>http://www.apache.org/licenses/LICENSE-2.0</a><br><br>
# Unless required by applicable law or agreed to in writing, software<br>
# distributed under the License is distributed on an "AS IS" BASIS,<br>
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.<br>
# See the License for the specific language governing permissions and<br>
# limitations under the License.<br>
# </p>
#

# %% [markdown]
# # ChipScoPy ILA Monitor Status Example
#
#
# <img src="../img/api_overview.png" width="500" align="left">

# %% [markdown]
# ## Description
# An advanced example showing how to monitor the ILA capture status, by running the function ILA.monitor_status(), in asynchroneous mode.
#
# By calling ChipScoPy ILA API function **monitor_status()**, in asynchronous mode we can do the following:
# - Get periodic updates of ILA core capture status.
# - Call other ChipScoPy API functions, while monitoring the ILA status.
# - Cancel the status monitor early.
#
#
# ## Requirements
# - Local or remote Xilinx Versal board, such as a VCK190
# - Xilinx hw_server 2021.2 installed and running
# - Xilinx cs_server 2021.2 installed and running
# - Python 3.8 or greater installed
# - ChipScoPy 2021.2 installed
# - Jupyter notebook support installed - Please do so, using the command `pip install chipscopy[jupyter]`

# %% [markdown]
# # Overview
# ### Function monitor_status() Documentation
#     A. Function monitor_status()
#     B. Calling monitor_status() in Synchronous Mode
#     C. Calling monitor_status() in Asynchronous Mode
#     D. Future Object
#     E. Progress Function
#     F. ILAStatus Class
#     G. ILAState Enum Values
# ### Example Steps
#     1.  Initialization: Imports and File Paths
#     2.  Create a session and connect to the hw_server and cs_server
#     3.  Program the Device
#     4.  Discover Debug Cores
#     5.  VIO Control and ILA Capture
#     6.  Configure the counter using VIO output probes
#     7.  Define Status Progress Function
#     8.  Arm ILA to Trigger on VIO Up/Down virtual switch
#     9.  Start Status Monitor
#     10. Toggle Counter Direction - Monitor Status
#     11. Status Monitor Completes
#     12. Upload and Print Data
# --------------------

# %% [markdown]
# ## A. Function monitor_status()
#
#     def monitor_status(
#        self, max_wait_minutes: float = None, progress=None, done: request.DoneFutureCallback = None
#     ) -> ILAStatus or request.CsFutureRequestSync:
#
# This function monitors ILA capture status and waits until all data has been captured, or until timeout
# or the function is cancelled.
# Call this function after arming the ILA and before uploading the waveform.
# The command operates in synchronous mode if *done* argument has default value *None* .
#
#
# Args:
# -  **max_wait_minutes** (float): Max time in minutes for status monitor. If *None*, status monitor never times out.
# -  **progress**(progress_fn) : See Asynchronous Mode, below. This function runs in the TCF Event Thread.
# -  **done**(request.DoneFutureCallback): Done callback. This function runs in the TCF Event Thread.
#
# Returns:
# - ILAStatus when called synchronously.
# - "future object" when called asynchronously.

# %% [markdown]
# ## B. Calling monitor_status() in Synchronous Mode
# - Typical mode when using this function. It is not covered in this example.
# - Function waits until all data has been captured in the ILA core, or timeout.
# - Use default argument value *None* for both arguments *progress* and *done* .
# - Returns an ILAStatus object.

# %% [markdown]
# ## C. Calling monitor_status() in Asynchronous Mode
#
# - This mode is useful for reporting the capture status to stdout or a GUI.
# - Function does not block. The main thread continues with the next statement.
# - Returns a *future* object, which represents the monitor.
# - A blocking function should be called later on the *future* object, in order to let the calling thread wait until the status monitor has completed.
# - Asynchronous Mode is selected, by specifying a *done* function, which is called after the function has completed.
# - If no user defined callback is needed, set *done* argument to dummy function *chipscopy.null_callback*, to enable asynchronous mode.

# %% [markdown]
# ## D. Future Object
#
# When the *monitor_status* function is called in asynchronous mode, it will return a *future* object.
# The *future* object has blocking attributes and functions, which will block the current thread until
# the status monitor has completed.
#
# ###  Blocking Attributes
#
# - *future.result* - Returns *None* or *ILAStatus*. *ILAStatus* object if capture did complete successfully, without timeout.
# - *future.error* - Returns *None* or *Exception*. *None* if no error otherwise an exception object, e.g. timeout exception.
#
# ### Non-blocking Attribute
#
# - *future.progress* - Returns *None* or *ILAStatus*. Read this attribute in the *progress function*, to get the ILA capture status.
#
# ### Blocking Function
#
# - *future.wait(timeout=None)* - *None* value means wait until status monitor completes. Argument *timeout* is in seconds.
#
# ### Non-blocking Function
#
# - *future.cancel()*  - Cancels the status monitor. An exception will be raised in the thread, which called the function *ILA.monitor_status*.

# %% [markdown]
# ## E. Progress Function
#
# - User-defined function which takes one argument *future*.
# - Useful for reporting ILA capture status.
# - The *progress function* is only called when the ILA capture status changes.
# - The *progress function* is called in the TCF Event Thread. The *progress function* must not call any ChipScoPy API function, which interacts with the cs_server or the device.
#
# >    Example:
# >
# >        def monitor_status_done(future):
# >            if not future.error:
# >                # future.result holds an ILAStatus object.
# >                print_status(future.progress)

# %% [markdown]
# ## F. ILAStatus Class
#
# The following data attributes are available to read, when monitoring the ILA capture status.
#
# | Attributes                  | TYPE     | Description                                                 |
# |:--------------------------- |:-------- |:----------------------------------------------------------- |
# | capture_state:              | ILAState | Capture state. See below.                                   |
# | is_armed                    | bool     | Trigger is armed.                                           |
# | is_full                     | bool     | Data buffer is full.                                        |
# | samples_captured            | int      | Number of samples captured in current data window.          |
# | windows_captured            | int      | Number of fully captured data windows.                      |
# | samples_requested           | int      | Requested number of samples per window, when was ILA armed. |
# | windows_requested           | int      | Requested number of windows, when ILA was armed.            |
# | trigger_position_requested  | int      | Requested trigger position, when ILA was armed.             |

# %% [markdown]
# ## G.  ILAState Enum Values
#
# ILA Capture State Transitions:<br>
# >        IDLE -> PRE_TRIGGER -> TRIGGER -> POST_TRIGGER -> IDLE
#
# | Enum Value   | Description
# |:-------------|:-------------------------------- |
# | IDLE         | Not armed.                       |
# | PRE_TRIGGER  | Capturing pre-trigger samples.   |
# | TRIGGER      | Waiting for trigger.             |
# | POST_TRIGGER | Capturing post-trigger samples.  |
#

# %% [markdown]
# # Example Steps

# %% [markdown]
# ## 1 - Initialization: Imports and File Paths
#
# After this step,
# - Required functions and classes are imported
# - URL paths are set correctly
# - File paths to example files are set correctly

# %%
import os
from time import sleep
import chipscopy
from chipscopy import get_design_files
from chipscopy import get_examples_dir_or_die, null_callback
from chipscopy import create_session, report_versions
from chipscopy.api.ila import export_waveform, get_waveform_data, ILAStatus, ILAState

# %%
# Specify locations of the running hw_server and cs_server below.
# To make things convenient, we default to values from the following environment variables.
# Modify these if needed.

CS_URL = os.getenv("CS_SERVER_URL", "TCP:localhost:3042")
HW_URL = os.getenv("HW_SERVER_URL", "TCP:localhost:3121")

# The get_design_files() function tries to find the PDI and LTX files. In non-standard
# configurations, you can put the path for PROGRAMMING_FILE and PROBES_FILE below.
design_files = get_design_files("vck190/production/chipscopy_ced")

PROGRAMMING_FILE = design_files.programming_file
PROBES_FILE = design_files.probes_file

print(f"HW_URL: {HW_URL}")
print(f"CS_URL: {CS_URL}")
print(f"PROGRAMMING_FILE: {PROGRAMMING_FILE}")
print(f"PROBES_FILE:{PROBES_FILE}")


# %% [markdown]
# ## 2 - Create a session and connect to the hw_server and cs_server
#
# The session is a container that keeps track of devices and debug cores.
# After this step,
# - Session is initialized and connected to server(s)
# - Versions are detected and reported to stdout

# %%
session = create_session(cs_server_url=CS_URL, hw_server_url=HW_URL)
report_versions(session)

# %% [markdown]
# ## 3 - Program the device with the example design
#
# After this step,
# - Device is programmed with the example programming file

# %%
# Typical case - one device on the board - get it.
device = session.devices.filter_by(family="versal").get()
device.program(PROGRAMMING_FILE)


# %% [markdown]
# ## 4 - Discover Debug Cores
#
# Debug core discovery initializes the chipscope server debug cores. This brings debug cores in the chipscope server online.
#
# After this step,
#
# - The cs_server is initialized and ready for use
# - ILA and VIO core instances in the device are reported

# %%
device.discover_and_setup_cores(ltx_file=PROBES_FILE)
print(f"Debug cores setup and ready for use.")

# %%
# Print out the ILA core instance UUIDs and instance names
ila_cores = device.ila_cores
for index, ila_core in enumerate(ila_cores):
    print(f"{index} - {ila_core.core_info.uuid}   {ila_core.name}")

# %%
# Print out the VIO core instance UUIDs and instance names
vio_cores = device.vio_cores
for index, vio_core in enumerate(vio_cores):
    print(f"{index} - {vio_core.core_info.uuid}   {vio_core.name}")


# %% [markdown]
# ## 5 - VIO Control and ILA Capture
#
# ILA and VIO are two important building blocks for debugging applications in hardware.
# This example design shows how to control IP using a VIO core and capture results with ILA.
#
# In this design,
# - A VIO core controls the counter (reset, up/down, ce, load)
# - An ILA core captures the counter values
#

# %% [markdown]
# <img src="img/capture_data.png" width="400" align="left">

# %%
# Grab the two cores we are interested in for the demonstration
# As shown above, a counter is connected to the ILA core.
# The VIO core controls the counter.

ila = device.ila_cores.get(name="chipscopy_i/counters/ila_slow_counter_0")
vio = device.vio_cores.get(name="chipscopy_i/counters/vio_slow_counter_0")

print(f"Using ILA: {ila.core_info.uuid}  {ila.name}")
print(f"Using VIO: {vio.core_info.uuid}  {vio.name}")


# %% [markdown]
# ## 6 - Configure the counter using VIO output probes
#
# Set up the VIO core to enable counting up, from 0
#
# <img src="img/vio_control_counter.png" width="300" align="left">

# %%
vio.reset_vio()
vio.write_probes(
    {
        "chipscopy_i/counters/slow_counter_0_SCLR": 0,
        "chipscopy_i/counters/slow_counter_0_L": 0x00000000,
        "chipscopy_i/counters/slow_counter_0_LOAD": 0,
        "chipscopy_i/counters/slow_counter_0_UP": 1,
        "chipscopy_i/counters/slow_counter_0_CE": 1,
    }
)
print("Counter is now free-running and counting up")


# %% [markdown]
# ## 7 - Define Status Progress Function
# - The **status_progress** function will be given to the **ila.monitor_status()** function, as an argument.
# - The function is called when status changes, but not more often than twice per second.
# - The argument **future** has a member *progress* of type **ILAStatus**, which contains the status information.
# - The **status_progress** function,
#     - prints status to stdout
#     - cancels the **monitor_status()** function, if 3 waveform windows have been captured.
# - Note: The monitor function will run in the Event Tread, which handles commands sent to the cs_server.
# - Important: Do not call an API function which communicates with the cs_server or device, from inside the monitor function. It may put the program in a deadlock.

# %%
def status_progress(future):
    """Called in Event Thread"""
    st: ILAStatus = future.progress
    if st:
        print_status(st)
        if st.windows_captured >= 3:
            print("\nCancelling ILA Status Monitor")
            future.cancel()


def print_status(st: ILAStatus):
    current_window = min(st.windows_captured + 1, st.windows_requested)
    print(
        f"\nCapture State: {st.capture_state.name} - "
        f" Windows Captured: {st.windows_captured} of {st.windows_requested}.",
        end="",
    )
    if st.is_full:
        print(f" - Capture is complete.")
    else:
        print(f" - Samples in current window: " f"{st.samples_captured} of {st.samples_requested}.")


# %% [markdown]
# ## 8 - Arm ILA to Trigger on VIO Up/Down virtual switch
#
# - Set ILA to trigger when UP/DOWN counter signal edge rises or falls.
# - Set ILA core to capture on a transition of the UP/DOWN toggle switch
# - Once transition happens, trigger in the middle of the buffer.
# - Request 10 windows.

# %% [markdown]
# <img src="img/edge_trigger.png" width="550" align="left">

# %%
ila.reset_probes()
ila.set_probe_trigger_value("chipscopy_i/counters/slow_counter_0_UP_1", ["==", "B"])
ila.run_basic_trigger(window_count=10, window_size=8, trigger_position=4)

print("ILA is armed")


# %% [markdown]
# ## 9 - Start Status Monitor
# - Define a *done* function *monitor_status_done*, which will be called when status monitoring ends.
# - Start the Status Monitor.
# - Passing in the *status_progress* function, defined above.
# - Get the *future* object from the *monitor_status* function.

# %%
def monitor_status_done(future):
    print("\nCallback function monitor_status_done() called.")
    if not future.error:
        # future.result holds an ILAStatus object.
        print_status(future.result)


# Called asynchronously, with done callback function.
# This thread will continue executing while monitoring of the ILA Status will happen in the event thread.
future = ila.monitor_status(
    max_wait_minutes=0.5, progress=status_progress, done=monitor_status_done
)

# %% [markdown]
# ## 10 - Toggle Counter Direction - Monitor Status
# - Use VIO to toggle counter up/down switch 3 times: DOWN/UP/DOWN.
# - This will cause the running ILA to trigger 3 times, capturing 3 windows of data.
# - The *status_progress* function will call *future.cancel()*, after 3 windows.
# - Observe messages:
#     - *status_progress* prints capture status.
#     - *status_progress* prints out message when calling *future.cancel()*.
#     - *done* function *monitor_status_done* prints message when called.

# %%
for switch_value in [0, 1, 0]:
    print("\nChanging counter up/down direction.")
    vio.write_probes({"chipscopy_i/counters/slow_counter_0_UP": switch_value})
    # Sleep 2.0 seconds.
    sleep(2.0)

# %% [markdown]
# ## 11 - Status Monitor Completes
# - *future.result* will block until capture completes or is cancelled.
# - Status monitoring is cancelled.
# - *future.result* thows the cancel exception.

# %%
status = None
try:
    # future.result is blocking, until monitor_status() function has completed or timed-out or been cancelled.
    status = future.result
except Exception as ex:
    # Catch cancel exception here.
    print(f"\n****** Status Monitor raised exception: '{ex}' ******\n")

if status:
    print("Function monitor_status() has completed.")
else:
    ila.refresh_status()

print(f"Windows avaliable to upload: {ila.status.windows_captured}.")

# %% [markdown]
# ## 12 - Upload and Print Data
# - Print the captured ILA samples and mark the trigger position.
# - Note that counter changes direction after the trigger mark.

# %%
ila.upload()
samples = get_waveform_data(
    ila.waveform,
    ["chipscopy_i/counters/slow_counter_0_Q_1"],
    include_trigger=True,
    include_sample_info=True,
)
for trigger, sample_index, window_index, window_sample_index, value in zip(*samples.values()):
    trigger = "<-- Trigger" if trigger else ""
    print(
        f"Window:{window_index}  Window Sample:{window_sample_index}  {value:10}  0x{value:08X} {trigger}"
    )



# %% [markdown]
# <link rel="preconnect" href="https://fonts.gstatic.com">
# <link href="https://fonts.googleapis.com/css2?family=Fira+Code&display=swap" rel="stylesheet">
#
# ### License
#
# <p style="font-family: 'Fira Code', monospace; font-size: 1.2rem">
# Copyright 2021 Xilinx, Inc.<br><br>
# Licensed under the Apache License, Version 2.0 (the "License");<br>
# you may not use this file except in compliance with the License.<br><br>
# You may obtain a copy of the License at <a href="http://www.apache.org/licenses/LICENSE-2.0"?>http://www.apache.org/licenses/LICENSE-2.0</a><br><br>
# Unless required by applicable law or agreed to in writing, software<br>
# distributed under the License is distributed on an "AS IS" BASIS,<br>
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.<br>
# See the License for the specific language governing permissions and<br>
# limitations under the License.<br>
# </p>
#

# %% [markdown]
# # ChipScoPy ILA and VIO Example
#
#
# <img src="../img/api_overview.png" width="500" align="left">

# %% [markdown]
# ## Description
# This example demonstrates how to program and communicate with ILA (Integrated Logic Analyzer) and
# VIO (Virtual IO) cores using the ChipScoPy Python API.
#
#
# ## Requirements
# - Local or remote Xilinx Versal board, such as a VCK190
# - Xilinx hw_server 2021.2 installed and running
# - Xilinx cs_server 2021.2 installed and running
# - Python 3.8 or greater installed
# - ChipScoPy 2021.2 installed
# - Jupyter notebook support installed - Please do so, using the command `pip install chipscopy[jupyter]`

# %% [markdown]
# ## 1 - Initialization: Imports and File Paths
#
# After this step,
# - Required functions and classes are imported
# - URL paths are set correctly
# - File paths to example files are set correctly

# %%
import sys
import os
from chipscopy import get_design_files
from chipscopy import create_session, report_versions
from chipscopy.api.ila import get_waveform_data, export_waveform

# %%
# Make sure to start the hw_server and cs_server prior to running.
# Specify locations of the running hw_server and cs_server below.
# The default is localhost - but can be other locations on the network.
CS_URL = os.getenv("CS_SERVER_URL", "TCP:localhost:3042")
HW_URL = os.getenv("HW_SERVER_URL", "TCP:localhost:3121")

# The get_design_files() function tries to find the PDI and LTX files. In non-standard
# configurations, you can put the path for PROGRAMMING_FILE and PROBES_FILE below.
design_files = get_design_files("vck190/production/chipscopy_ced")

PROGRAMMING_FILE = design_files.programming_file
PROBES_FILE = design_files.probes_file

print(f"HW_URL: {HW_URL}")
print(f"CS_URL: {CS_URL}")
print(f"PROGRAMMING_FILE: {PROGRAMMING_FILE}")
print(f"PROBES_FILE:{PROBES_FILE}")


# %% [markdown]
# ## 2 - Create a session and connect to the hw_server and cs_server
#
# The session is a container that keeps track of devices and debug cores.
# After this step,
# - Session is initialized and connected to server(s)
# - Versions are detected and reported to stdout

# %%
session = create_session(cs_server_url=CS_URL, hw_server_url=HW_URL)
report_versions(session)

# %% [markdown]
# ## 3 - Program the device with the example design
#
# After this step,
# - Device is programmed with the example programming file

# %%
# Typical case - one device on the board - get it.
device = session.devices.filter_by(family="versal").get()
device.program(PROGRAMMING_FILE)


# %% [markdown]
# ## 4 - Discover Debug Cores
#
# Debug core discovery initializes the chipscope server debug cores. This brings debug cores in the chipscope server online.
#
# After this step,
#
# - The cs_server is initialized and ready for use
# - ILA and VIO core instances in the device are reported

# %%
device.discover_and_setup_cores(ltx_file=PROBES_FILE)
print(f"Debug cores setup and ready for use.")

# %%
# Print out the ILA core instance UUIDs and instance names
ila_cores = device.ila_cores
for index, ila_core in enumerate(ila_cores):
    print(f"{index} - {ila_core.core_info.uuid}   {ila_core.name}")

# %%
# Print out the VIO core instance UUIDs and instance names
vio_cores = device.vio_cores
for index, vio_core in enumerate(vio_cores):
    print(f"{index} - {vio_core.core_info.uuid}   {vio_core.name}")


# %% [markdown]
# ## 5 - VIO Control and ILA Capture
#
# ILA and VIO are two important building blocks for debugging applications in hardware.
# This example design design shows how to control IP using a VIO core and capture results with ILA.
#
# In this Design,
# - A VIO core controls the counter (reset, up/down, ce, load)
# - An ILA core captures the counter values
#

# %% [markdown]
# <img src="img/capture_data.png" width="400" align="left">

# %%
# Grab the two cores we are interested in for the demonstration
# As shown above, a counter is connected to the ILA core.
# The VIO core controls the counter.

ila = device.ila_cores.get(name="chipscopy_i/counters/ila_slow_counter_0")
vio = device.vio_cores.get(name="chipscopy_i/counters/vio_slow_counter_0")

print(f"Using ILA: {ila.core_info.uuid}  {ila.name}")
print(f"Using VIO: {vio.core_info.uuid}  {vio.name}")


# %% [markdown]
# ### 5a - Configure the counter using VIO output probes
#
# <img src="img/vio_control_counter.png" width="300" align="left">

# %%
# Print all the VIO port and probe names. This is convenient to know which probes are connected to
# VIO ports. Also verifies probe names to pass to other functions.

print("VIO Port <---> Probe mapping:")
for probe in vio.probes:
    if probe.direction == "in":
        print(f"{probe.port_name} <-- {probe.probe_name}")
    else:
        print(f"{probe.port_name} --> {probe.probe_name}")

# %%
# Set up the VIO core to enable counting up from 0
#
vio.reset_vio()
vio.write_probes(
    {
        "chipscopy_i/counters/slow_counter_0_SCLR": 0,
        "chipscopy_i/counters/slow_counter_0_L": 0x00000000,
        "chipscopy_i/counters/slow_counter_0_LOAD": 0,
        "chipscopy_i/counters/slow_counter_0_UP": 1,
        "chipscopy_i/counters/slow_counter_0_CE": 1,
    }
)
print("Counter is now free-running and counting up")

# %% [markdown]
# ### 5b - Capture and display free-running counter using the ILA core
#
# <img src="img/free_running_counter.png" width="350" align="left">

# %%
# Trigger ILA on the free running counter. Trigger set to the first time we see 0s in low 16-bits.
# This will show the counter is free running, and counting up

ila.reset_probes()
ila.set_probe_trigger_value("chipscopy_i/counters/slow_counter_0_Q_1", ["==", "0xXXXX_0000"])
ila.run_basic_trigger(window_count=1, window_size=32, trigger_position=16)
print("ILA is running - looking for trigger")

# %%
# Wait for the ILA trigger with upload.
# Then print the captured ILA samples and mark the trigger position.

ila.wait_till_done(max_wait_minutes=0.1)
upload_successful = ila.upload()
if upload_successful:
    samples = get_waveform_data(
        ila.waveform,
        ["chipscopy_i/counters/slow_counter_0_Q_1"],
        include_trigger=True,
        include_sample_info=True,
    )
    for trigger, sample_index, window_index, window_sample_index, value in zip(*samples.values()):
        trigger = "<-- Trigger" if trigger else ""
        print(
            f"Window:{window_index}  Window Sample:{window_sample_index}  {value:10}  0x{value:08X} {trigger}"
        )
else:
    print("Failed to upload ILA data from core")

# %% [markdown]
# ### 5c - Trigger ILA using VIO Up/Down virtual switch
#
# This step demonstrates how VIO and ILA can be combined to form powerful debug building blocks.
#
# ILA is set to trigger when UP/DOWN counter signal edge rises or falls.
# VIO drives the UP/DOWN counter control signal to 0 causing the counter to count down.
# The signal transition causes ILA to trigger and capture data.
#
# After this step,
# - VIO drives counter to count from UP to DOWN
# - ILA triggers on the UP to DOWN signal transition
# - Waveform uploaded with the up/down trigger sample in the center of captured data
#

# %% [markdown]
# <img src="img/edge_trigger.png" width="550" align="left">

# %%
# Set ILA core to capture on a transition of the UP/DOWN toggle switch
# Once transition happens, trigger in the middle of the buffer.

ila.reset_probes()
ila.set_probe_trigger_value("chipscopy_i/counters/slow_counter_0_UP_1", ["==", "B"])
ila.run_basic_trigger(window_count=1, window_size=32, trigger_position=16)

print("ILA is running - looking for trigger")

# %%
# VIO: Turn counter up/down switch to DOWN position.
# This will cause the running ILA to trigger on the transition edge from up to down.

vio.write_probes({"chipscopy_i/counters/slow_counter_0_UP": 0})

print("VIO changed up/down counter to count down")

# %%
# Print the captured ILA samples and mark the trigger position.
# Note that counter counts down after the trigger mark.

ila.wait_till_done(max_wait_minutes=0.1)
upload_successful = ila.upload()
if upload_successful:
    samples = get_waveform_data(
        ila.waveform,
        ["chipscopy_i/counters/slow_counter_0_Q_1"],
        include_trigger=True,
        include_sample_info=True,
    )
    for trigger, sample_index, window_index, window_sample_index, value in zip(*samples.values()):
        trigger = "<-- Trigger" if trigger else ""
        print(
            f"Window:{window_index}  Window Sample:{window_sample_index}  {value:10}  0x{value:08X} {trigger}"
        )
else:
    print("Failed to upload ILA data from core")

# %% [markdown]
# ## 6 - Waveform Export - VCD (or CSV)
#
#  Demonstrate how to export waveform data to a VCD file for visualizing waveform in other tools.
#
#  Export includes complete waveform with probe, _TRIGGER, and _WINDOW.
#
#  - For CSV export, substitute "CSV" for "VCD" argument.
#  - To export to a file, substitute the filename for 'sys.stdout'

# %%
if upload_successful:
    export_waveform(ila.waveform, "VCD", sys.stdout)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Generator, Tuple

from chipscopy import dm
from chipscopy.dm import request
from chipscopy.proxies.DebugCoreProxy import DebugCoreService
from chipscopy.tcf.channel import Token


class DebugCoreClientNode(dm.Node):
    """
    This node makes use of the DebugCore service.
    """

    @staticmethod
    def is_compatible(node: dm.Node) -> bool:
        return not node.parent_ctx

    def read_bytes(
        self,
        address: int,
        data: bytearray,
        offset: int = 0,
        byte_size: int = None,
        done: request.DoneCallback = None,
        progress: request.ProgressCallback = None,
    ) -> Token:
        """
        Reads from memory context, using DebugCore Service.

        Args:
            address (int): Start memory address, to read from.
            data (bytearray): Buffer for return data.
            offset (int): Start byte index *data* buffer, where return data is copied to.
            byte_size (int): Number of byte to read. Default is the size of *data* buffer.
            done (request.DoneCallback):
            progress (request.ProgressCallback):

        Returns: Token
        """

        done = request._make_callback(done)
        initial_token = None
        local_pending = set()
        all_requests_sent = False

        total_read_size = len(data) if byte_size is None else byte_size
        if total_read_size > len(data) - offset:
            raise IndexError("Request to read more data than buffer can hold, from {self.ctx}.")

        chunk_byte_size = 0x8000
        service = self.manager.channel.getRemoteService(DebugCoreService)
        err = None
        seq_it = None

        def get_request_sequence() -> Generator[Tuple[Dict[str, Any], int], None, None]:
            end_addr = address + total_read_size
            for addr, buffer_offset in zip(
                range(address, end_addr, chunk_byte_size),
                range(offset, offset + total_read_size, chunk_byte_size),
            ):
                r_size = chunk_byte_size if addr + chunk_byte_size < end_addr else end_addr - addr
                seq = {"name": "read bytes", "type": "r", "addr": addr, "size": r_size // 4}
                yield seq, buffer_offset

        def request_data() -> Token:
            nonlocal all_requests_sent
            return_token = None

            if err:
                read_done()
                return

            while len(local_pending) < 8:
                seq, buffer_offset = next(seq_it, (None, None))
                if not seq:
                    all_requests_sent = True
                    break
                else:
                    token = service.run_sequence(self.ctx, [seq], receive_data)
                    local_pending.add(token)
                    if not return_token:
                        return_token = token
                    token.data_offset = buffer_offset

            return return_token

        def receive_data(token, error, result):
            nonlocal err

            local_pending.remove(token)
            if err:
                return

            err = error
            if (
                not error
                and hasattr(token, "data_offset")
                and isinstance(result, list)
                and len(result) == 1
                and isinstance(result[0], dict)
            ):
                r_data = result[0]["data"]
                start = token.data_offset
                end = start + len(r_data)
                if end > offset + total_read_size:
                    err = IndexError(
                        "Request to read more data than buffer can hold, from {self.ctx}."
                    )
                else:
                    data[start:end] = r_data
                    if progress:
                        progress((end - offset) / total_read_size)

            if all_requests_sent and not local_pending:
                read_done()
            else:
                request_data()

        def read_done():
            nonlocal initial_token
            token = initial_token
            initial_token = None
            if done:
                done.done_request(token, err, None)

        #
        seq_it = get_request_sequence()
        initial_token = request_data()
        return initial_token

    def write_bytes(
        self,
        address: int,
        data: bytearray,
        offset: int = 0,
        byte_size: int = None,
        done: request.DoneCallback = None,
        progress: request.ProgressCallback = None,
    ) -> Token:
        """
        Read from memory context, using DebugCore Service.

        Args:
            address (int): Start memory address, to write to.
            data (bytearray): Buffer with write data.
            offset (int): Start byte index *data*.
            byte_size (int): Number of byte to write. Default is the size of *data* buffer.
            done (request.DoneCallback):
            progress (request.ProgressCallback):

        Returns: Token
        """

        done = request._make_callback(done)
        initial_token = None
        local_pending = set()
        all_sent = False

        total_write_size = len(data) if byte_size is None else byte_size
        if total_write_size > len(data) - offset:
            raise IndexError("Request to write more data than data buffer size, to {self.ctx}.")

        chunk_byte_size = 0x8000
        service = self.manager.channel.getRemoteService(DebugCoreService)
        err = None
        seq_it = None

        def get_send_sequence() -> Generator[Tuple[Dict[str, Any], int], None, None]:
            end_addr = address + total_write_size
            for addr, buffer_offset in zip(
                range(address, end_addr, chunk_byte_size),
                range(offset, offset + total_write_size, chunk_byte_size),
            ):
                w_size = chunk_byte_size if addr + chunk_byte_size < end_addr else end_addr - addr
                buf = data[buffer_offset : buffer_offset + w_size]
                seq = {"name": "write bytes", "type": "w", "addr": addr, "data": buf}
                yield seq, buffer_offset + w_size

        def send_data() -> Token:
            nonlocal all_sent
            return_token = None

            if err:
                write_done()
                return

            while len(local_pending) < 8 and not all_sent:
                seq, buffer_next_offset = next(seq_it, (None, None))
                if seq:
                    token = service.run_sequence(self.ctx, [seq], send_confirmation)
                    local_pending.add(token)
                    if not return_token:
                        return_token = token
                    token.data_offset = buffer_next_offset
                else:
                    all_sent = True
                    break

            return return_token

        def send_confirmation(token, error, _):
            nonlocal err
            nonlocal all_sent

            local_pending.remove(token)
            if err:
                return

            err = error
            if not error and hasattr(token, "data_offset"):
                buffer_next_offset = token.data_offset
                if progress:
                    progress((buffer_next_offset - offset) / total_write_size)

            if local_pending or not all_sent:
                send_data()
            else:
                write_done()

        def write_done():
            nonlocal initial_token
            token = initial_token
            initial_token = None
            if done:
                done.done_request(token, err, None)

        #
        seq_it = get_send_sequence()
        initial_token = send_data()
        return initial_token



# *****************************************************************************
# * Copyright (c) 2011, 2013 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *****************************************************************************

"""
ChannelProxy implements forwarding of TCF messages between two channels.
The class is used to implement Locator service "redirect" command.
"""

from .. import channel


class ProxyCommandListener(channel.CommandListener):
    def __init__(self, ch, tokens):
        self.ch = ch
        self.tokens = tokens

    def progress(self, token, data):
        self.ch.sendProgress(self.tokens.get(token), data)

    def result(self, token, data):
        self.ch.sendResult(self.tokens.pop(token, None), data)

    def terminated(self, token, error):
        self.ch.rejectCommand(self.tokens.pop(token, None))


class ChannelProxy(object):
    def __init__(self, x, y):
        # assert not isinstance(x, ChannelLoop)
        # assert not isinstance(y, ChannelLoop)
        self.ch_x = x
        self.ch_y = y
        assert self.ch_x.getState() == channel.STATE_OPEN
        assert self.ch_y.getState() == channel.STATE_OPENING
        self.tokens_x = {}
        self.tokens_y = {}
        cmd_listener_x = ProxyCommandListener(self.ch_x, self.tokens_x)
        cmd_listener_y = ProxyCommandListener(self.ch_y, self.tokens_y)
        proxy = self

        class ProxyX(channel.Proxy):
            def onChannelClosed(self, error):
                proxy.closed_x = True
                if proxy.closed_y:
                    return
                if error is None:
                    proxy.ch_y.close()
                else:
                    proxy.ch_y.terminate(error)

            def onCommand(self, token, service, name, data):
                if proxy.closed_y:
                    return
                assert proxy.ch_y.getState() == channel.STATE_OPEN
                s = proxy.ch_y.getRemoteService(service)
                if not s:
                    proxy.ch_x.terminate(IOError("Invalid service name"))
                else:
                    key = proxy.ch_y.sendCommand(s, name, data, cmd_listener_x)
                    proxy.tokens_x[key] = token

            def onEvent(self, service, name, data):
                s = proxy.ch_x.getRemoteService(service)
                if not s:
                    proxy.ch_x.terminate(IOError("Invalid service name"))
                elif not proxy.closed_y:
                    proxy.ch_y.sendEvent(s, name, data)

        class ProxyY(channel.Proxy):
            def onChannelClosed(self, error):
                proxy.closed_y = True
                if proxy.closed_x:
                    return
                if error is None:
                    proxy.ch_x.close()
                else:
                    proxy.ch_x.terminate(error)

            def onCommand(self, token, service, name, data):
                if proxy.closed_x:
                    return
                assert proxy.ch_x.getState() == channel.STATE_OPEN
                s = proxy.ch_x.getRemoteService(service)
                if not s:
                    proxy.ch_y.terminate(IOError("Invalid service name"))
                else:
                    key = proxy.ch_x.sendCommand(s, name, data, cmd_listener_y)
                    proxy.tokens_y[key] = token

            def onEvent(self, service, name, data):
                s = proxy.ch_y.getRemoteService(service)
                if not s:
                    proxy.ch_y.terminate(IOError("Invalid service name"))
                elif not proxy.closed_x:
                    proxy.ch_x.sendEvent(s, name, data)

        proxy_x = ProxyX()
        proxy_y = ProxyY()

        try:
            self.ch_y.setProxy(proxy_y, self.ch_x.getRemoteServices())

            class ChannelListener(channel.ChannelListener):
                def onChannelClosed(self, error):
                    proxy.ch_y.removeChannelListener(self)
                    if error is None:
                        error = Exception("Channel closed")

                def onChannelOpened(self):
                    proxy.ch_y.removeChannelListener(self)
                    try:
                        proxy.ch_x.setProxy(proxy_x,
                                            proxy.ch_y.getRemoteServices())
                    except IOError as e:
                        proxy.ch_x.terminate(e)
                        proxy.ch_y.terminate(e)
            self.ch_y.addChannelListener(ChannelListener())
        except IOError as e:
            self.ch_x.terminate(e)
            self.ch_y.terminate(e)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Union, Callable, Any
from chipscopy.tcf.services import DoneHWCommand, Service
from chipscopy.client.core_property_client import PropertyValues
from chipscopy.client.hbm_mc_client import HBMMCClient
from chipscopy.proxies.CorePropertyProxy import CorePropertyProxy

# Memory Models
MEMORY_MODEL_HBM = "memory_model_hbm"
HBMMS_DOMAIN_INDEX = 1

# Service name.
SERVICE_NAME = "HBM"


class HBMService(Service):
    """TCF HBM Service interface."""

    def getName(self) -> str:
        """
        Get service name of this service.

        :returns: Service name :const:`NAME`.
        """
        return SERVICE_NAME


class HBMProxy(CorePropertyProxy, HBMService):
    def __init__(self, channel):
        super().__init__(channel)
        self._listeners = {}

    #
    #  General methods
    #
    def initialize(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_command("initialize", (node_id,), done)

    def write32(
        self, node_id: str, write_addr: int, data: int, domain_index: int, done: DoneHWCommand
    ) -> None:
        """
        Write 32 bit register.

        :param node_id: Core context id.
        :param write_addr: Address.
        :param data: 32 bit data
        :param domain_index: HBMMS only (1)
        """
        return self.send_command("write32", (node_id, write_addr, data, domain_index), done)

    def read32(
        self,
        node_id: str,
        read_addr: int,
        read_word_count: int,
        domain_index: int,
        done: DoneHWCommand,
    ) -> Union[List[int], int]:
        """
        Read 32 bit register

        :param node_id: Core context id.
        :param read_addr: address to read from.
        :param read_word_count: number of 32 bit words.
        :param domain_index: HBMMS only (1)
        :return int value if read_word_count == 1. Otherwise list of int values.
        """
        return self.send_command(
            "read32", (node_id, read_addr, read_word_count, domain_index), done
        )

    def refresh_property(
        self, node_id: str, prop_names: List[str], done: DoneHWCommand
    ) -> PropertyValues:
        return self.send_command("refreshProperty", (node_id, prop_names), done)

    def terminate(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_command("terminate", (node_id,), done)

    #
    #  Service specific user interfaces
    #
    def get_enabled_mcs_mask(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_command("getEnabledMcsMask", (node_id,), done)

    def get_init_status(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_command("getInitStatus", (node_id,), done)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Iterable, Any, Dict, NewType, Callable
from chipscopy import dm
from chipscopy.dm import request
from chipscopy.tcf.services import xicom
from chipscopy.proxies.XicomProxy import XicomProxy as XicomService
from chipscopy.proxies.JtagProxy import JtagProxy as JtagService
from chipscopy.proxies.JtagCableProxy import JtagCableProxy as JtagCableService
from chipscopy.proxies.JtagDeviceProxy import JtagDeviceProxy as JtagDeviceService
from chipscopy.tcf.protocol import invokeLater
from chipscopy.utils.logger import log


class DoneCmd(xicom.DoneXicom):
    def __init__(self, node, done):
        self.node = node
        self.done = done

    def doneHW(self, token, error, props):
        self.node.update(props)
        self.node.remove_pending(token)  # done state change
        self.done.done_request(token, error, props)


class JtagRegister(object):
    def __init__(self, node: "JtagDevice", reg_ctx: str, definition: Dict):
        self.reg_ctx = reg_ctx
        self.definition = definition
        self.data = None
        self.fields = {}
        self.node = node

    def update(self, done: request.DoneCallback = None):
        proc = self.node.manager.channel.getRemoteService(XicomService)
        done = request._make_callback(done)

        def done_update(token, error, results):
            if not error:
                self.data = results
                self.parse_fields()
            if done:
                done.done_request(token, error, results)

        return proc.jtag_reg_get(self.node.ctx, self.reg_ctx, done_update)

    @property
    def is_static(self):
        return self.definition.get("is_static", False)

    def __str__(self):
        return f"{self.reg_ctx}({self.definition['length']} bits)"

    def __repr__(self):
        return self.__str__()

    def __getitem__(self, item):
        return self.fields[item]["value"]

    def __getattr__(self, item):
        field = self.fields.get(item)
        return field["value"] if field else None

    def parse_fields(self):
        self.fields = {}
        fields = self.definition.get("fields")
        if fields:
            for field in fields:
                bits = field.get("bits")
                if bits:
                    data_field = {"name": field["name"]}
                    if len(bits) > 1:
                        data_field["bit_range"] = f"{bits[0]}:{bits[-1]}"
                    else:
                        data_field["bit_range"] = f"{bits[0]}"
                    if self.data:
                        value = 0
                        for b in bits:
                            bit = (self.data[int(b / 8)] >> (b % 8)) & 1
                            value = (value << 1) | bit
                        data_field["value"] = value
                    enums = field.get("enums")
                    if enums:
                        data_enums = {}
                        for e in enums:
                            data_enums[e["value"]] = e
                        data_field["enums"] = data_enums
                    self.fields[field["name"]] = data_field

    def print_reg(self):
        hex_str = ""
        if self.data:
            hex_str = "0x" + self.data[::-1].hex()
        print(f"{self.definition['name']} = {hex_str}")

    def print_fields(self):
        self.print_reg()
        max_name_len = 0
        for field in self.fields.values():
            max_name_len = max(max_name_len, len(field["name"]) + len(field["bit_range"]) + 12)
        for field in self.fields.values():
            label = f"{field['name']} (Bits [{field['bit_range']}])"
            value = field["value"]
            if "enums" in field and value in field["enums"]:
                print(f"{label.rjust(max_name_len, ' ')}: {field['enums'][value]['name']}")
            else:
                print(f"{label.rjust(max_name_len, ' ')}: {field['value']}")


DoneGetRegister = NewType("DoneGetRegister", Callable[[Any, Exception or None, JtagRegister], None])


class JtagRegDefManager(object):
    def __init__(self):
        self.reg_defs = {}

    def get_registers(self, node: "JtagDevice", done: DoneGetRegister):
        """
        Gets available registers for a given JTAG device.  The definitions are supplied but data is not set.

        :param node:  JTAG device for which to get registers
        :param done: Done callback
        :return: Token of request
        """
        proc = node.manager.channel.getRemoteService(XicomService)
        assert proc
        ret_token = None
        pending = set()
        regs = {}

        def check_done(error):
            nonlocal done
            if (not pending or error) and done:
                done(ret_token, error, regs)
                done = None

        def done_list(token, error, reg_list):
            if not error:
                for reg in reg_list:
                    if reg not in self.reg_defs:
                        pending.add(proc.jtag_reg_def(reg, handle_done_def(reg)))
                    else:
                        regs[self.reg_defs[reg]["name"]] = JtagRegister(
                            node, reg, self.reg_defs[reg]
                        )
            check_done(error)

        def handle_done_def(reg_ctx):
            def done_def(token, error, reg_def):
                pending.remove(token)
                if not error:
                    self.reg_defs[reg_ctx] = reg_def
                    regs[reg_def["name"]] = JtagRegister(node, reg_ctx, reg_def)
                check_done(error)

            return done_def

        # add pending to indicate that this node is being changed
        ret_token = proc.jtag_reg_list(node.ctx, done_list)
        return ret_token


class JtagDevice(dm.Node):
    @staticmethod
    def is_compatible(node: dm.Node) -> bool:
        return bool(node.isTap) and node.parent_ctx != ""

    def post_init(self):
        self._update_reg_defs()
        self._update_device_properties()

    @staticmethod
    def get_def_name(reg_name: str):
        """
        Get register bit definition key from a register name
        :param reg_name: register name
        :return: register bit defifinition key
        """
        return reg_name + "_definition"

    def update_regs(
        self, reg_names: Iterable[str] = [], force: bool = False, done: request.DoneCallback = None
    ):
        """
        Updates the register values for all registers that need updating.

        :param reg_names: List of reg_names to update. Empty list means update all that need updating.
        :param force: Forces all registers to update
        :param done: Done callback
        :return: Token of request
        """
        done = request._make_callback(done)
        ret_token = None

        def done_update(token, error, results):
            nonlocal done
            self.remove_pending(token)
            if (self.is_ready or error) and done:
                done.done_request(ret_token, error, self["regs"])
                done = None

        # Start updates for all registers that don't have data already or aren't static
        if reg_names:
            for name in reg_names:
                reg = self["regs"].get(name)
                if reg:
                    ret_token = self.add_pending(reg.update(done_update))
        else:
            for reg in self["regs"].values():
                if force or (not reg.data or not reg.is_static):
                    ret_token = self.add_pending(reg.update(done_update))

        if not ret_token:
            invokeLater(done_update, None, Exception("No registers to update"), None)
        return ret_token

    def config(self, filepath: str, props: Dict[str, Any] = None):
        """
        Configures the device with a given PDI file
        :param filepath: File path of the PDI
        :param props: Configuration properties
        :param done: DoneRequest callback
        :param progress: Progress callback
        :return: Token of request
        """
        assert self.request
        proc = self.manager.channel.getRemoteService(XicomService)
        err = None
        f = open(filepath, "rb")
        total_size = os.stat(filepath).st_size
        total_sent = 0
        total_progress = 0
        chunk_size = 0x4000

        def send_data():
            nonlocal total_sent
            if err:
                config_done(None, err, None)
                return
            while len(self.pending) < 8 and total_sent < total_size:
                data = f.read(chunk_size)
                tok = self.add_pending(proc.config_data(self.ctx, data, config_progress))
                tok.data_size = len(data)
                total_sent += len(data)
                if total_sent >= total_size:
                    self.add_pending(proc.config_end(self.ctx, props, config_wait_for_dpc_ready))

        def config_started(token, error, results):
            nonlocal err
            self.remove_pending(token)
            err = error
            send_data()

        def config_progress(token, error, results):
            nonlocal err
            nonlocal total_progress
            self.remove_pending(token)
            if not err:
                err = error
                if err:
                    config_done(None, err, None)
                    return
                if hasattr(token, "data_size"):
                    total_progress += token.data_size
                self.request.set_progress(total_progress / total_size)
                send_data()

        def config_wait_for_dpc_ready(token, error, results):
            nonlocal err
            if token:
                self.remove_pending(token)
            if not err:
                err = error
            # Wait for DPC to come online by forcing get_children
            # Not doing this can result in a timing issue and timeout exception
            # on subsequent dpc memory read.
            dc_proc = self.manager.channel.getRemoteService("DebugCore")
            assert dc_proc
            self.add_pending(dc_proc.get_children("", config_done))

        def config_done(token, error, results):
            nonlocal err
            if token:
                self.remove_pending(token)

            f.close()
            if not err:
                err = error

            if err:
                self.request.set_exception(err)
            else:
                self.request.set_result(None)

        # add pending to indicate that this node is being changed
        self.add_pending(proc.config_begin(self.ctx, {}, config_started))
        send_data()

    def reset(self):
        """
        Performs a soft reset (SYS_RST IR command)
        :param done: DoneRequest callback
        :return: Token of request
        """
        assert self.request
        proc = self.manager.channel.getRemoteService(XicomService)

        def done_reset(token, error, results):
            self.remove_pending(token)
            if error:
                self.set_exception(error)
            else:
                self.request.set_result(results)

        # add pending to indicate that this node is being changed
        self.add_pending(proc.config_reset(self.ctx, {}, done_reset))

    def status(self, reg_name: str = "jtag_status", done: request.DoneCallback = None):
        """
        Updates and prints the given register.

        :param reg_name: Name of the register
        :param done: Done callback
        :return: Token of request
        """
        done = request._make_callback(done)

        def done_update(token, error, result):
            if not error:
                self.regs[reg_name].print_fields()
            if done:
                done.done_request(token, error, self.regs[reg_name])

        self.update_regs((reg_name,), True, done_update)

    def secure_debug(self, filepath: str, done: request.DoneCallback = None):
        """
        Enables secure debug on a locked device.
        :param filepath: Path to authentication file
        :param done: DoneRequest callback
        :return: Token of request
        """
        proc = self.manager.channel.getRemoteService(XicomService)
        done = request._make_callback(done)
        f = open(filepath, "rb")
        data = f.read()

        def done_run(token, error, results):
            print(f"Done secure debug {token.getID()}")
            self.remove_pending(token)
            if done:
                done.done_request(token, error, results)

        # add pending to indicate that this node is being changed
        return self.add_pending(proc.secure_debug(self.ctx, data, done_run))

    def readback(
        self,
        filepath: str = "",
        readback_word_count: int = 0,
        outfilepath: str = "",
        done: request.DoneCallback = None,
    ):
        """
        Reads back config data.  Requires a PDI to set up the readback through SBI.
        :param filepath: Path to CDO readback PDI
        :param readback_word_count: Optional argument to set the number of words to readback.  If not set it will
        attempt to determine readback size during reading.
        :param outfilepath:  Filepath to output file if desired
        :param done: DoneRequest callback
        :return: Token of request
        """
        proc = self.manager.channel.getRemoteService(XicomService)
        ret_token = None

        def done_run(token, error, results):
            self.remove_pending(token)
            if outfilepath and results:
                with open(outfilepath, "wb") as f:
                    f.write(results)
            if done:
                done(ret_token, error, results)

        def done_config(token, error, results):
            self.remove_pending(token)
            props = {}
            if error and "DONE" in str(error):
                error = None
            if not error:
                if readback_word_count:
                    props["size"] = readback_word_count * 4
                from time import sleep

                sleep(0.5)  # delay to give the readback pdi time to load data into SBI
                return self.add_pending(proc.readback_config(self.ctx, props, done_run))
            elif done:
                done(ret_token, error, results)
            return token

        config_props = {"skip_done_check": True}
        # add pending to indicate that this node is being changed
        if filepath:
            ret_token = self.add_pending(self.config(filepath, config_props, done_config))
        else:
            ret_token = done_config(None, None, None)
        return ret_token

    def _update_reg_defs(self):
        def done_update_reg_defs(token, error, regs):
            self.remove_pending(token)
            if not error:
                if regs:
                    self["regs"] = regs

        self.add_pending(jtag_reg_defs.get_registers(self, done_update_reg_defs))

    def _update_device_properties(self):
        jtag_device = self.manager.channel.getRemoteService(JtagDeviceService)

        def done_get_properties(token, error, props):
            if not error:
                if props:
                    self.update(props)
            self.remove_pending(token)

        self.add_pending(jtag_device.get_properties(self.idCode, done_get_properties))


class JtagCable(dm.Node):
    @staticmethod
    def is_compatible(node: dm.Node) -> bool:
        return node.parent_ctx == "" and node.Name == "whole scan chain"

    def post_init(self):
        self._update_static_properties()
        self.get_option("frequency")
        self.get_option("frequency_list")
        self.get_option("skew")
        self.get_option("timing_class")

    def set_freq(self, freq: int, done: request.DoneCallback = None):
        """ Sets the frequency of the cable """
        proc = self.manager.channel.getRemoteService(JtagService)
        assert proc
        ret_token = None

        def done_get_freg(token, error, results):
            if done:
                done(ret_token, error, results)

        def done_set_freq(token, error, results):
            self.remove_pending(token)
            self.get_option("frequency", done_get_freg)

        ret_token = self.add_pending(proc.set_option(self.ctx, "frequency", freq, done_set_freq))
        return ret_token

    def get_option(self, key: str, done: request.DoneCallback = None):
        """
        Retrieves the frequency of the cable from the hw_server

        :param key: Key of the option
        :param done: Request callback when complete
        """
        proc = self.manager.channel.getRemoteService(JtagService)
        assert proc

        def done_get_option(token, error, results):
            self.remove_pending(token)
            self[key] = results
            if done:
                done(token, error, results)

        return self.add_pending(proc.get_option(self.ctx, key, done_get_option))

    def _update_static_properties(self, done: request.DoneCallback = None):
        proc = self.manager.channel.getRemoteService(JtagCableService)
        assert proc

        def done_update_static(token, error, results):
            self.remove_pending(token)
            self.update(results)
            if done:
                done(token, error, results)

        return self.add_pending(proc.getContext(self.ctx, done_update_static))


jtag_reg_defs = JtagRegDefManager()



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import shutil
import os
import argparse
import sys
from distutils.dir_util import mkpath, copy_tree
from distutils.file_util import copy_file
from shutil import move
import chipscopy

SOURCE_EXAMPLE_DIR = os.path.abspath(os.path.join(os.path.dirname(chipscopy.__file__), "examples"))
TARGET_EXAMPLE_DIR = os.path.join(".", "chipscopy-examples")


def _copy_and_move_files(files_to_copy, files_to_move):
    """Copy and move files and folders. ``files_to_copy`` and ``files_to_move``
    are expected to be dict where the key is the source path, and the value is
    destination path.
    """
    # copy files and folders
    for src, dst in files_to_copy.items():
        if os.path.isfile(src):
            mkpath(os.path.dirname(dst))
            copy_file(src, dst)
        else:
            copy_tree(src, dst)
    # and move files previously downloaded
    for src, dst in files_to_move.items():
        shutil.move(src, dst)


def deliver_examples(src_path, dst_fullpath, name):
    """Deliver examples to target destination path.

    Parameters
    ----------
        src_path: str
            The source path to copy from
        dst_fullpath: str
            The destination path to copy to
        name: str
            The name of the example module
    """
    files_to_copy = {}
    files_to_move = {}
    for root, dirs, files in os.walk(src_path):
        if ".ipynb_checkpoints" in dirs:
            dirs.remove(".ipynb_checkpoints")
        if "__init__.py" in files:
            files.remove("__init__.py")
        keepers = [".ipynb", ".py", ".pdi", ".ltx", ".png"]
        for f in files:
            extension = os.path.splitext(f)[1]
            if extension in keepers:
                # If folder is in the list of files to copy, remove it as it is
                # going to be inspected
                if root in files_to_copy:
                    files_to_copy.pop(root)
                relpath = os.path.relpath(root, src_path)
                relpath = "" if relpath == "." else relpath
                try:
                    files_to_copy_tmp = {}
                    files_to_move_tmp = {}
                    for d in dirs:
                        if d != "__pycache__":
                            dir_dst_path = os.path.join(dst_fullpath, relpath, d)
                            files_to_copy_tmp[os.path.join(root, d)] = dir_dst_path
                    for f2 in files:
                        file_dst_path = os.path.join(dst_fullpath, relpath, f2)
                        files_to_copy_tmp[os.path.join(root, f2)] = file_dst_path
                    files_to_copy.update(files_to_copy_tmp)
                    files_to_move.update(files_to_move_tmp)
                except FileNotFoundError as e:  # pragma: no cover
                    if relpath:
                        nb_str = os.path.join(name, relpath)
                        print(
                            "Could not resolve file '{}' in folder "
                            "'{}', examples will not be "
                            "delivered".format(str(e), nb_str)
                        )
    try:
        if not files_to_copy:
            print("The example module '{}' could not be delivered. ".format(name))
        else:
            _copy_and_move_files(files_to_copy, files_to_move)
    except (Exception, KeyboardInterrupt) as e:  # pragma: no cover
        print("Exception detected. Delivery process did not complete...")
        raise e


def find_examples(example_dir=SOURCE_EXAMPLE_DIR, full_path=False):
    examples_list = []
    for root, dirs, files in os.walk(example_dir):
        if ".ipynb_checkpoints" in dirs:
            dirs.remove(".ipynb_checkpoints")
        if "__pycache__" in dirs:
            dirs.remove("__pycache__")
        if "__init__.py" in files:
            files.remove("__init__.py")
        examples = [f for f in files if f.endswith(".ipynb") or f.endswith(".py")]
        if full_path:
            examples = [os.path.join(root, f) for f in examples]
        examples_list.extend(examples)
    return examples_list


class _GetExamplesParser(argparse.ArgumentParser):  # pragma: no cover
    @property
    def epilog(self):
        return "Available example modules: {}".format(", ".join(find_examples()))

    @epilog.setter
    def epilog(self, x):
        """Dont set epilog in Parser.__init__."""
        pass


def _get_examples_parser():  # pragma: no cover
    """Initialize and return the argument parser."""
    parser = _GetExamplesParser(description="Deliver available ChipScoPy examples")
    parser.add_argument(
        "-l", "--list", action="store_true", help="List available examples and exit"
    )
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Force delivery even if target examples "
        "directory already exists. The existing "
        "directory will be renamed adding a timestamp",
    )
    parser.add_argument(
        "-p",
        "--path",
        type=str,
        help="Specify a custom path to deliver examples to. "
        "Default is '{}'".format(TARGET_EXAMPLE_DIR),
    )
    parser.add_argument(
        "-y", "--assume_yes", action="store_true", help='Skip Interactive Prompts - assume "Yes"'
    )
    return parser


def main():  # pragma: no cover
    parser = _get_examples_parser()
    args = parser.parse_args()
    examples_list = find_examples()
    if args.list:
        if examples_list:
            print("Available examples:\n- {}".format("\n- ".join(examples_list)))
        else:
            print("No examples available")
            sys.exit(2)
    if not examples_list:
        print("No examples available, nothing can be delivered")
        sys.exit(2)
    if args.path:
        delivery_path = args.path
    else:
        delivery_path = TARGET_EXAMPLE_DIR
    yes = ["yes", "ye", "y"]
    no = ["no", "n"]
    nb_str = "\n- ".join([os.path.basename(nb) for nb in examples_list])
    print(
        f"The following examples  will be delivered to {os.path.abspath(delivery_path)}: \n- {nb_str}"
    )
    if not args.assume_yes:
        choice = input("Do you want to proceed? [Y/n] ").lower()
        while True:
            if choice == "" or choice in yes:
                break
            if choice in no:
                return
            choice = input("Please respond with 'yes' or 'no' (or 'y' or " "'n') ")
    if os.path.exists(delivery_path):
        if args.force:
            import datetime

            timestamp = datetime.datetime.now().strftime("%Y_%m_%d_%H%M%S")
            backup_dir = os.path.split(delivery_path)[1] + "_" + timestamp
            backup_fullpath = os.path.join(os.path.dirname(delivery_path), backup_dir)
            move(delivery_path, backup_fullpath)
            print(f"Backing up existing examples to {backup_dir}")
        else:
            print(
                "Target examples directory already "
                "exists. Specify another path or use "
                "the 'force' option to proceed"
            )
            sys.exit(1)
    try:
        for example_name in examples_list:
            print(f"Delivering example '{os.path.basename(example_name)}'...")
            deliver_examples(SOURCE_EXAMPLE_DIR, delivery_path, example_name)
    except (Exception, KeyboardInterrupt) as e:
        raise e
    finally:
        if os.path.isdir(delivery_path) and len(os.listdir(delivery_path)) == 0:
            os.rmdir(delivery_path)
        if not os.path.isdir(delivery_path):
            print("No examples available")
            sys.exit(2)

    sys.exit(0)


if __name__ == "__main__":  # pragma: no cover
    main()



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
ChipScoPy report generation. This module keeps stdout reporting in one place for convenience.

All reporting functions that generate stdout reports should be here with a function prefixed with "report_"
hand have a meaningful function name in this file. Other modules should call this one to do the reporting.

Should we want to replace rich in the future, most rich reporting and formatting code is conveniently in one place.

Don't add stdout prints in other locations. This is an API to interact with hardware so we want stdout to
 be optional and well controlled.

The idea is to have the ChipScoPy API generate data that can pipe into other functions.
"""

import datetime
import json
from typing import Optional, Union, List, TYPE_CHECKING

from rich.table import Table
from rich.box import SQUARE as BOX_SQUARE

import chipscopy
from chipscopy.utils.printer import printer
from chipscopy.utils.version import ServerVersionInfo

if TYPE_CHECKING:
    from chipscopy.api.device.device import Device
    from chipscopy.api.session import Session


def _create_server_report(server_version: ServerVersionInfo):
    """Generate a pretty report for the connected server status."""
    report = Table(box=BOX_SQUARE)
    report.add_column("Attribute", justify="right")
    report.add_column("Value", justify="left")
    report.add_row("Status", "Connected")
    report.add_row("Build", f"{server_version.build}")
    report.add_row("Version", f"{server_version.version}")
    report.add_row("Timestamp", f"{server_version.timestamp}")
    if server_version.server_type == "cs_server":
        report.add_row("Package", f"{server_version.package}")
        report.add_row("Artifact type", f"{server_version.artifact}")
    return report


def report_versions(session: Optional["Session"] = None):
    """
    Generate a version report for the servers and ChipScoPy API.

    Args:
        session: Instance of :py:class:`~chipscopy.api.session.Session`

    """
    report = Table(title="ChipScoPy Version Information", box=BOX_SQUARE)
    report.add_column("Entity", justify="right")
    report.add_column("Version", justify="left")

    chipscopy_report = Table(box=BOX_SQUARE)
    chipscopy_report.add_column("Attribute", justify="right")
    chipscopy_report.add_column("Value", justify="left")
    chipscopy_report.add_row("Build", chipscopy.__version__)
    chipscopy_report.add_row(
        "Timestamp",
        datetime.datetime.fromtimestamp(int(chipscopy.__version__.split(".")[-1])).strftime(
            "%b %d %Y-%H:%M:%S"
        ),
    )

    report.add_row("ChipScoPy", chipscopy_report)

    if session and session.hw_server:
        hw_server_version_info = ServerVersionInfo(session.hw_server, "hw_server")
        hw_server_report = _create_server_report(hw_server_version_info)
        report.add_row("", "")
        report.add_row(
            f"hw_server @ {session.hw_server.url}",
            hw_server_report,
        )

    if session and session.cs_server:
        cs_server_version_info = ServerVersionInfo(session.cs_server, "cs_server")
        cs_server_report = _create_server_report(cs_server_version_info)
        report.add_row("", "")
        report.add_row(f"cs_server @ {session.cs_server.url}", cs_server_report)

    printer("\n")
    printer(report)
    printer("\n")


def _create_device_report(device: "Device", title="ChipScoPy Device"):
    report = Table(title=title, box=BOX_SQUARE)
    json_data = json.loads(repr(device.device_identification))
    report.add_column("Property", justify="left")
    report.add_column("Value", justify="left")
    for key in ["family", "dna", "cable_name", "jtag_index"]:
        report.add_row(key, str(json_data[key]))
    for target_dict in json_data["node_identification"]:
        report.add_row(target_dict["hier_name"], target_dict["context"])
    return report


def report_devices(arg: Optional[Union[List["Device"], "Device", "Session"]] = None):
    from chipscopy.api.device.device import Device
    from chipscopy.api.session import Session

    devices = []
    if isinstance(arg, Session):
        devices = arg.devices
    elif isinstance(arg, Device):
        # convert to a list
        devices = [arg]
    for device in devices:
        sub_report = _create_device_report(device, "Device report")
        printer("\n")
        printer(sub_report)
        printer("\n")


# TODO: Add other report_* functions in this file for report generation.
# TODO: report_devices(dna=123456, ...)
# TODO: report_targets(jtag=true, memory=true, ...)
# TODO: report_cores(...)


def report_hierarchy(obj: Union["IBERT"]):
    """
    Reports the child object hierarchy for any given parent object.

    .. note: Not all objects support reporting hierarchy. No output will be generated for
        such objects.

    Args:
        obj: Parent object

    """
    try:
        printer(obj.__rich_tree__())
    except AttributeError:
        pass



# *****************************************************************************
# * Copyright (c) 2020 Xilinx, Inc.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Xilinx
# *****************************************************************************

import threading
from .. import compat
from .StreamChannel import StreamChannel, ESC


class CircularQueue(object):
    def __init__(self, maxsize):
        self.queue = bytearray(maxsize)
        self.head = 0
        self.tail = 0
        self.sent = 0
        self.last_byte = 0
        self.eom_bytes = (1, 2, 3)
        self.maxsize = maxsize
        self.mutex = threading.Lock()
        self.not_empty = threading.Condition(self.mutex)
        self.not_full = threading.Condition(self.mutex)

    def _next(self, i):
        return (i + 1) % self.maxsize

    def full(self):
        return self._next(self.tail) == self.head

    def empty(self):
        return self.head == self.sent

    def put(self, data):
        update_empty = self.empty()

        tail = self._next(self.tail)
        if tail == self.head:
            with self.not_full:
                self.not_full.wait()

        self.queue[tail] = data
        self.tail = tail

        if self.last_byte == ESC and data in self.eom_bytes:
            self.sent = tail

            if update_empty:
                with self.not_empty:
                    self.not_empty.notify()

        self.last_byte = data
        return True

    def wait_for_not_empty(self):
        with self.not_empty:
            self.not_empty.wait()

    def get(self):
        update_full = self.full()

        if self.head == self.sent:
            with self.not_empty:
                self.not_empty.wait()

        head = self._next(self.head)
        data = self.queue[head]
        self.head = head

        if update_full:
            with self.not_full:
                self.not_full.notify()

        return data

    def size(self):
        if self.tail >= self.head:
            return self.tail - self.head
        return self.maxsize - (self.head - self.tail)


class LoopbackChannel(StreamChannel):
    def __init__(self, remote_peer, local_peer=None):
        super(LoopbackChannel, self).__init__(remote_peer, local_peer)
        self.closed = False
        self.queue = CircularQueue(0x1000)
        self.start()

    def getBuf(self, buf):
        if self.closed:
            return -1
        if self.queue.empty():
            self.queue.wait_for_not_empty()
        count = min(self.queue.size(), len(buf))
        for n in range(count):
            buf[n] = self.queue.get()
        return count

    def get(self):
        if self.closed:
            return -1
        return self.queue.get()

    @staticmethod
    def str2bytes(data):
        if isinstance(data, compat.strings):
            return bytearray([ord(x) for x in data])
        elif isinstance(data, int):
            return bytearray([data])
        return data

    def put(self, b):
        if self.closed:
            return
        s = self.str2bytes(b)
        for p in s:
            self.queue.put(p)

    def putBuf(self, buf):
        if self.closed:
            return
        if isinstance(buf, (bytes, bytearray)):
            s = buf
        else:
            s = self.str2bytes(buf)
        for p in s:
            self.queue.put(p)

    def flush(self):
        pass

    def stop(self):
        self.closed = True



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List
from chipscopy import dm
from chipscopy.client import connect_xicom
from chipscopy.client.server_info import ServerInfo
from chipscopy.client.hbm_mc_client import HBMMCClient
from chipscopy.client.core import get_cs_view, CoreParent
from chipscopy.client.core_property_client import CorePropertyClient
from chipscopy.tcf import protocol
from chipscopy.tcf.services import DoneHWCommand
from chipscopy.dm.request import DoneCallback
from chipscopy.proxies.HBMProxy import SERVICE_NAME, HBMMS_DOMAIN_INDEX


class HBMClient(CorePropertyClient):
    @staticmethod
    def is_compatible(node: dm.Node) -> bool:
        return node.type and node.type == "hbm"

    def get_service_proxy(self):
        return self.manager.channel.getRemoteService(SERVICE_NAME)

    #
    #  General methods
    #
    def initialize(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        return self.add_pending(service.initialize(self.ctx, done_cb))

    def write32(self, write_addr: int, data: int, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.write32(self.ctx, write_addr, data, HBMMS_DOMAIN_INDEX, done_cb)
        return self.add_pending(token)

    def read32(self, read_addr: int, read_word_count: int, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.read32(self.ctx, read_addr, read_word_count, HBMMS_DOMAIN_INDEX, done_cb)
        return self.add_pending(token)

    def refresh_property(self, names: List[str], done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.refresh_property(self.ctx, names, done_cb)
        return self.add_pending(token)

    def terminate(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        return self.add_pending(service.terminate(self.ctx, done_cb))

    #
    #  Service specific user interfaces
    #
    def get_hbm_mc(self, index: int, done: DoneCallback):
        mc_count = len(self.children)
        if (index < 0) or (index + 1 == mc_count):
            raise Exception("Invalid HBM MC index given.")

        mc_context = self.children[index]
        hbm_mc = self.manager.get_node(mc_context, HBMMCClient)
        done(result=hbm_mc)
        return

    def get_enabled_mcs_mask(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.get_enabled_mcs_mask(self.ctx, done_cb)
        return self.add_pending(token)

    def get_init_status(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.get_init_status(self.ctx, done_cb)
        return self.add_pending(token)


def get_hbm(server: ServerInfo, hbm_index=0) -> HBMClient:

    cs_view = get_cs_view(server)
    dpc = None
    for node in cs_view.get_children():
        if "xcvc1902" in node.props.get("Name"):
            dpc = cs_view.get_node(node.ctx, CoreParent)
            break

    # Find HBM stack under DPC with a given HBM index
    hbm = server.cs_target(parent=dpc, type="hbm", index=hbm_index, cls=HBMClient)

    return hbm


def get_hbm_mc(server: ServerInfo, hbm_stack: HBMClient, mc_index=0) -> HBMMCClient:

    # Find HBM MC under specific HBM Stack with a given MC index
    hbm_mc = server.cs_target(parent=hbm_stack, type="hbm_mc", index=mc_index, cls=HBMMCClient)

    return hbm_mc



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from chipscopy.tcf.services import Service, DoneHWCommand
from chipscopy.proxies.CorePropertyProxy import CorePropertyProxy

NOC_PERFMON_SERVICE_NAME = "NoCPerfMon"


class NoCPerfMonService(Service):
    def getName(self) -> str:
        return NOC_PERFMON_SERVICE_NAME


class NoCPerfMonProxy(CorePropertyProxy, NoCPerfMonService):
    def __init__(self, channel):
        super(NoCPerfMonProxy, self).__init__(channel)
        self.listeners = {}

    def initialize(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_xicom_command("initialize", (node_id,), done)

    def discover_noc_elements(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_xicom_command("discoverNOCElements", (node_id,), done)

    def configure_monitors(self, params, done: DoneHWCommand) -> None:
        return self.send_xicom_command("configureMonitors", (params["node_id"], params), done)

    def get_supported_sampling_periods(self, params, done: DoneHWCommand):
        return self.send_xicom_command(
            "getSupportedSamplingPeriods", (params["node_id"], params), done
        )

    def get_clk_info(self, node_id: str, done: DoneHWCommand):  # pragma: no cover
        return self.send_xicom_command("getClkInfo", (node_id,), done)

    def enumerate_noc_elements(self, params, done: DoneHWCommand):
        return self.send_xicom_command("enumerateNocElements", (params["node_id"], params), done)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from chipscopy.api.ibert.ibert import IBERT
from chipscopy.api.ibert.link.manager import LinkGroupManager, LinkManager
from chipscopy.api.ibert.eye_scan.manager import EyeScanManager
from chipscopy.api.ibert.yk_scan.manager import YKScanManager

# Aliases for ease of use
create_links = LinkManager.create_links
delete_links = LinkManager.delete_links
get_all_links = LinkManager.all_links

create_link_groups = LinkGroupManager.create_link_groups
delete_link_groups = LinkGroupManager.delete_link_groups
get_all_link_groups = LinkGroupManager.all_link_groups

create_eye_scans = EyeScanManager.create_eye_scans
delete_eye_scans = EyeScanManager.delete_eye_scans
get_all_eye_scans = EyeScanManager.all_eye_scans

create_yk_scans = YKScanManager.create_yk_scans
delete_yk_scans = YKScanManager.delete_yk_scans
get_all_yk_scans = YKScanManager.all_yk_scans



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from chipscopy.tcf.services import Service, DoneHWCommand
from chipscopy.proxies.CorePropertyProxy import CorePropertyProxy
from chipscopy.utils.logger import log

AXIS_PCIE_NODE_NAME = "pcie"
AXIS_PCIE_SERVICE_NAME = "AxisPCIe"

DOMAIN_NAME = "proxy_pcie"


class AxisPCIeService(Service):
    def getName(self) -> str:
        return AXIS_PCIE_SERVICE_NAME


class AxisPCIeProxy(CorePropertyProxy, AxisPCIeService):
    def __init__(self, channel):
        super(AxisPCIeProxy, self).__init__(channel)
        self.listeners = {}

    def initialize(self, node_id: str, done: DoneHWCommand) -> None:
        log[DOMAIN_NAME].debug("Sending initializeCmd")
        return self.send_xicom_command("initialize", (node_id,), done)

    def read_data(self, node_id: str, done: DoneHWCommand) -> bool:
        return self.send_xicom_command("read_data", (node_id,), done)

    def reset_core(self, node_id: str, done: DoneHWCommand) -> bool:
        return self.send_xicom_command("reset_core", (node_id,), done)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from typing import Dict, List, Any, Union
from chipscopy.dm.request import DoneCallback


class CorePropertyProxy(ABC):
    def get_property(
        self, node_id: str, property_names: List[str], done: DoneCallback
    ) -> Dict[str, Any]:
        return self.send_xicom_command("getProperty", (node_id, property_names), done)

    def get_property_group(
        self, node_id: str, groups: List[str], done: DoneCallback
    ) -> Dict[str, Any]:
        return self.send_xicom_command(
            "getPropertyGroup",
            (
                node_id,
                groups,
            ),
            done,
        )

    def report_property(
        self, node_id: str, property_names: List[str], done: DoneCallback
    ) -> Dict[str, Dict[str, Any]]:
        return self.send_xicom_command(
            "reportProperty",
            (
                node_id,
                property_names,
            ),
            done,
        )

    def report_property_group(
        self, node_id: str, groups: List[str], done: DoneCallback
    ) -> Dict[str, Dict[str, Any]]:
        return self.send_xicom_command(
            "reportPropertyGroup",
            (
                node_id,
                groups,
            ),
            done,
        )

    def reset_property(self, node_id: str, property_names: List[str], done: DoneCallback):
        return self.send_xicom_command(
            "resetProperty",
            (
                node_id,
                property_names,
            ),
            done,
        )

    def reset_property_group(
        self, node_id: str, property_group_names: List[str], done: DoneCallback
    ):
        return self.send_xicom_command(
            "resetPropertyGroup",
            (
                node_id,
                property_group_names,
            ),
            done,
        )

    def set_property(self, node_id: str, property_values: Dict[str, Any], done: DoneCallback):
        return self.send_xicom_command(
            "setProperty",
            (
                node_id,
                property_values,
            ),
            done,
        )

    def set_property_group(self, node_id: str, property_values: Dict[str, Any], done: DoneCallback):
        return self.send_xicom_command(
            "setPropertyGroup",
            (
                node_id,
                property_values,
            ),
            done,
        )

    def list_property_groups(self, node_id: str, done: DoneCallback) -> list:
        return self.send_xicom_command("listPropertyGroups", (node_id,), done)

    def commit_property_group(self, node_id: str, groups: List[str], done: DoneCallback):
        return self.send_xicom_command(
            "commitPropertyGroup",
            (
                node_id,
                groups,
            ),
            done,
        )

    def refresh_property_group(
        self, node_id: str, groups: List[str], done: DoneCallback
    ) -> Dict[str, Any]:
        return self.send_xicom_command("refreshPropertyGroup", (node_id, groups), done)

    def add_to_property_watchlist(
        self, node_id: str, property_names: Union[str, List[str]], *, done: DoneCallback = None
    ):
        return self.send_xicom_command("add_to_property_watchlist", (node_id, property_names), done)

    def remove_from_property_watchlist(
        self, node_id: str, property_names: Union[str, List[str]], *, done: DoneCallback = None
    ):
        return self.send_xicom_command(
            "remove_from_property_watchlist", (node_id, property_names), done
        )

    def commit_memory(
        self,
        node_id: str,
        property_name: str,
        data: bytearray,
        start_byte_index: int,
        word_byte_length: int,
        done: DoneCallback = None,
    ) -> None:
        return self.send_xicom_command(
            "commitMemory", (node_id, property_name, data, start_byte_index, word_byte_length), done
        )

    def refresh_memory(
        self,
        node_id: str,
        property_name: str,
        byte_count: int,
        start_byte_index: int,
        word_byte_length: int,
        done: DoneCallback = None,
    ) -> None:
        return self.send_xicom_command(
            "refreshMemory",
            (node_id, property_name, byte_count, start_byte_index, word_byte_length),
            done,
        )



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Dict, Tuple, Union, Any
from itertools import islice

from chipscopy import dm
from chipscopy.client import connect_xicom
from chipscopy.client.core_property_client import PropertyValues
from chipscopy.client.core import get_cs_view, CoreParent, CoreClient
from chipscopy.tcf.services import DoneHWCommand, mila
from chipscopy.tcf.services.mila import (
    NAME as MILA_NAME,
    MILAProbeDefs,
    PROP_TRACE_DATA,
    PROP_TRACE_WIDTH,
)


class MILACoreClient(CoreClient):
    @staticmethod
    def is_compatible(node: dm.Node) -> bool:
        return node.type and node.type == "ddrmc_main"

    def get_service_proxy(self):
        return self.manager.channel.getRemoteService(MILA_NAME)

    #
    #  General methods
    #
    def initialize(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        return self.add_pending(service.initialize(self.ctx, done_cb))

    def terminate(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        return self.add_pending(service.terminate(self.ctx, done_cb))

    def write32(self, write_addr: int, data: int, domain_index: int, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.write32(self.ctx, write_addr, data, domain_index, done_cb)
        return self.add_pending(token)

    def read32(
        self, read_addr: int, read_word_count: int, domain_index: int, done: DoneHWCommand = None
    ):
        service, done_cb = self.make_done(done)
        token = service.read32(self.ctx, read_addr, read_word_count, domain_index, done_cb)
        return self.add_pending(token)

    #
    #  Property methods
    #
    def get_property(self, property_names: List[str], done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.get_property(self.ctx, property_names, done_cb)
        return self.add_pending(token)

    def get_property_group(self, groups: List[str], done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.get_property_group(self.ctx, groups, done_cb)
        return self.add_pending(token)

    def report_property(self, property_names: List[str], done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.report_property(self.ctx, property_names, done_cb)
        return self.add_pending(token)

    def report_property_group(self, groups: List[str], done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.report_property_group(self.ctx, groups, done_cb)
        return self.add_pending(token)

    def reset_property(self, property_names: List[str], done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.reset_property(self.ctx, property_names, done_cb)
        return self.add_pending(token)

    def reset_property_group(self, groups: List[str], done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.reset_property_group(self.ctx, groups, done_cb)
        return self.add_pending(token)

    def set_property(self, property_values: PropertyValues, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.set_property(self.ctx, property_values, done_cb)
        return self.add_pending(token)

    def commit_property_group(self, groups: List[str], done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.commit_property_group(self.ctx, groups, done_cb)
        return self.add_pending(token)

    def refresh_property_group(self, groups: List[str], done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.refresh_property_group(self.ctx, groups, done_cb)
        return self.add_pending(token)

    #
    # Probe Methods
    #
    def define_probe(self, probe_defs: mila.MILAProbeDefs, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.define_probe(self.ctx, probe_defs, done_cb)
        return self.add_pending(token)

    def undefine_probe(self, probe_names: List[str], done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.undefine_probe(self.ctx, probe_names, done_cb)
        return self.add_pending(token)

    def get_probe(self, probe_names: List[str], done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.get_probe(self.ctx, probe_names, done_cb)
        return self.add_pending(token)

    def get_probe_match_value(self, probe_names: List[str], done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.get_probe_match_value(self.ctx, probe_names, done_cb)
        return self.add_pending(token)

    def set_probe_match_value(self, match_pairs: Dict[str, str], done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.set_probe_match_value(self.ctx, match_pairs, done_cb)
        return self.add_pending(token)

    #
    # Core Communication Methods
    #
    def arm(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.arm(self.ctx, done_cb)
        return self.add_pending(token)

    def upload(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.upload(self.ctx, done_cb)
        return self.add_pending(token)

    #
    #  MUX Methods
    #
    def mux_add_node(
        self,
        memory_domain: str,
        addr: Union[int, None],
        name: str,
        long_name: str,
        has_local_inputs: bool = True,
        ch_in0: str = "",
        ch_in1: str = "",
        done: DoneHWCommand = None,
    ):
        service, done_cb = self.make_done(done)
        token = service.mux_add_node(
            self.ctx,
            memory_domain,
            addr,
            name,
            long_name,
            has_local_inputs,
            ch_in0,
            ch_in1,
            done_cb,
        )
        return self.add_pending(token)

    def mux_build_tree(
        self, clk_sel_is_1_mux: str, ila_capture_data_sel_muxes: [str], done: DoneHWCommand = None
    ):
        service, done_cb = self.make_done(done)
        token = service.mux_build_tree(
            self.ctx, clk_sel_is_1_mux, ila_capture_data_sel_muxes, done_cb
        )
        return self.add_pending(token)

    def mux_select_inputs(self, mux_node_inputs, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        # For JSON change tuples into lists.
        mux_node_inputs2 = [None if not item else list(item) for item in mux_node_inputs]
        token = service.mux_select_inputs(self.ctx, mux_node_inputs2, done_cb)
        return self.add_pending(token)

    def mux_get_selected_inputs(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done, restore_tuple=True)
        token = service.mux_get_selected_inputs(self.ctx, done_cb)
        return self.add_pending(token)

    def mux_commit(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.mux_commit(self.ctx, done_cb)
        return self.add_pending(token)

    def mux_refresh(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.mux_refresh(self.ctx, done_cb)
        return self.add_pending(token)

    def mux_report(self, skip_default: bool, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.mux_report(self.ctx, skip_default, done_cb)
        return self.add_pending(token)


#
#  Wavedrom Utilities
#
def wave_one_probe_to_int(tdata: bytearray, sample_width: int, slice: [int]) -> [int]:
    values = []
    msb, lsb = slice
    bytes_per_sample = (sample_width + 7) // 8
    for sample_low_byte in range(0, len(tdata), bytes_per_sample):
        val = 0
        for bit in range(msb, lsb - 1, -1):
            val = val << 1
            val += 1 if tdata[sample_low_byte + bit // 8] & (1 << (bit % 8)) else 0
        values.append(val)
    return values


def wave_to_int(probe_defs: MILAProbeDefs, wave_dict: Dict) -> Dict:
    """ Break up waveform by probe. Convert to list of int values per probe. """
    tdata = wave_dict[PROP_TRACE_DATA]
    sample_width = wave_dict[PROP_TRACE_WIDTH]
    wave_ints = {
        pname: wave_one_probe_to_int(tdata, sample_width, slice)
        for pname, slice in probe_defs.items()
    }
    return wave_ints


def to_wavedrom_wave_data_color(signal: Dict, start: int, end: int, signal_idx: int) -> None:
    # color 3: yellow, 4: orange, 5: blue, '=': no color
    color = "5" if (signal_idx % 4) == 0 else "="
    vals = signal["int_vals"]
    wave = []
    data = []
    for idx in range(start, end):
        if (idx > start) and (vals[idx] == vals[idx - 1]):
            wave.append(".")
        else:
            wave.append(color)
            data.append(f"{vals[idx]:x}")

    if data:
        signal["data"].append(" ".join(data))
    signal["wave"].append("".join(wave))
    return signal


def to_wavedrom_wave_data(signal: Dict, start: int, end: int) -> None:
    is_one_bit = signal["msb"] == signal["lsb"]
    vals = signal["int_vals"]
    wave = []
    data = []
    for idx in range(start, end):
        if (idx > start) and (vals[idx] == vals[idx - 1]):
            wave.append(".")
        elif is_one_bit:
            wave.append("0" if vals[idx] == 0 else "1")
        else:
            wave.append("=")
            data.append(f"{vals[idx]:x}")

    if data:
        signal["data"].append(" ".join(data))
    signal["wave"].append("".join(wave))
    return signal


def to_wave_json_str(
    signals: [Dict], display_count: int, display_width: int, start_sample: int, do_narrow=True
) -> str:
    def data_str(sig: Dict, index):
        if not sig.get("data", None) or len(sig["data"]) <= index:
            return ""

        return f', data: "{sig["data"][index]}"'

    narrow_str = ", config: {skin:'narrow'}" if do_narrow else ""

    res = ""
    for idx in range(0, display_count):
        head = '<script type="WaveDrom">\n{ signal: [\n'
        tick_start = start_sample + (idx * display_width)
        section_header = f', text: "Samples {tick_start} to {tick_start+display_width-1}"'
        tail = (
            "],\n head: { tick:"
            + str(tick_start)
            + section_header
            + "}"
            + narrow_str
            + " \n"
            + "}\n</script>\n"
        )
        body = [
            " { " + f'name: "{sig["name"]}", wave: "{sig["wave"][idx]}"' + data_str(sig, idx) + "},"
            for sig in signals
        ]
        section_str = head + "\n".join(body) + tail
        res += section_str
    return res


def make_name_with_range(name: str, msb: int, lsb: int) -> str:
    return f"{name}{[msb]}" if msb == lsb else f"{name}[{msb}:{lsb}]"


def wave_to_wave_json_multi(probe_defs: MILAProbeDefs, wave_dict: Dict, display_width: int) -> str:
    wave_ints = wave_to_int(probe_defs, wave_dict)
    signals = [
        {
            "name": make_name_with_range(name, probe_defs[name][0], probe_defs[name][1]),
            "int_vals": ints,
            "wave": [],
            "data": [],
            "msb": probe_defs[name][0],
            "lsb": probe_defs[name][1],
        }
        for name, ints in wave_ints.items()
    ]

    # Break up into sections:
    sample_count = wave_dict["trace_sample_count"]
    start_sample = -wave_dict["trace_trigger_position"]
    display_count = (sample_count + (display_width - 1)) // display_width
    for start in range(0, sample_count, display_width):
        signals_with_wave_data = [
            to_wavedrom_wave_data(signal, start, start + display_width) for signal in signals
        ]
    return to_wave_json_str(signals_with_wave_data, display_count, display_width, start_sample)


def wave_to_wave_json(
    probe_defs: MILAProbeDefs, wave_dict: Dict, center_sample: int = 0, sample_count: int = 64
) -> str:
    wave_ints = wave_to_int(probe_defs, wave_dict)
    signals = [
        {
            "name": make_name_with_range(name, probe_defs[name][0], probe_defs[name][1]),
            "int_vals": ints,
            "wave": [],
            "data": [],
            "msb": probe_defs[name][0],
            "lsb": probe_defs[name][1],
        }
        for name, ints in wave_ints.items()
    ]

    total_sample_count = wave_dict["trace_sample_count"]
    sample_count = min(total_sample_count, sample_count)
    trig_pos_list = wave_dict["trace_trigger_position"]
    trig_pos = trig_pos_list[0]
    start_idx = trig_pos + center_sample - sample_count // 2
    if start_idx < 0:
        start_idx = 0
    elif start_idx + sample_count >= total_sample_count:
        start_idx = total_sample_count - sample_count

    signals_with_wave_data = [
        to_wavedrom_wave_data_color(signal, start_idx, start_idx + sample_count, signal_idx)
        for signal_idx, signal in enumerate(signals)
    ]
    return to_wave_json_str(
        signals_with_wave_data, 1, sample_count, start_idx - trig_pos, do_narrow=sample_count > 50
    )


def display_waveform_in_browser(
    probe_defs: MILAProbeDefs,
    wave_dict: Dict,
    wavedrom_dir: str = ".",
    center_sample: int = 0,
    sample_count: int = 64,
) -> None:
    import webbrowser
    import os

    # create wave json
    wave_data = wave_to_wave_json(probe_defs, wave_dict, center_sample, sample_count)
    filename = "waveform.html"
    f = open(filename, "w")

    message = (
        """
    <html>
      <head>
        <title>Mini-ILA Waveform</title>
      </head>

        <body onload="WaveDrom.ProcessAll()">

        """
        + wave_data
        + f'<script src="{wavedrom_dir}/wavedrom/skins/default.js" type="text/javascript"></script>\n'
        + f'<script src="{wavedrom_dir}/wavedrom/skins/narrow.js" type="text/javascript"></script>\n'
        + f'<script src="{wavedrom_dir}/wavedrom/wavedrom.min.js" type="text/javascript"></script>\n'
        + """

        <hr>
      </body>
    </html>
    """
    )

    f.write(message)
    f.close()

    # Change path to reflect file location
    full_filename = "file:///" + os.getcwd() + "/" + filename
    webbrowser.open_new_tab(full_filename)


def display_waveform_in_jupyter(
    probe_defs: MILAProbeDefs,
    wave_dict: Dict,
    wavedrom_dir: str = ".",
    center_sample: int = 0,
    sample_count: int = 32,
) -> None:
    """ Show wave json waveform in Jupytor notebock cell"""
    import os
    import IPython

    def _is_javascript_present(javascript_path):
        return os.path.isfile(javascript_path)

    # create wave json
    wave_data = wave_to_wave_json(probe_defs, wave_dict, center_sample, sample_count)

    current_path = os.getcwd()
    wavedrom_min_js = os.path.join(current_path, "wavedrom", "wavedrom.min.js")
    wavedrom_skin_js = os.path.join(current_path, "wavedrom", "skins", "default.js")
    wavedrom_skin_narrow_js = os.path.join(current_path, "wavedrom", "skins", "narrow.js")

    if not (
        _is_javascript_present(wavedrom_skin_js)
        and _is_javascript_present(wavedrom_min_js)
        and _is_javascript_present(wavedrom_skin_narrow_js)
    ):
        print("paths", wavedrom_min_js, wavedrom_skin_js, sep="\n")
        raise Exception("wavedrom js not found")

    htmldata = " <h1>Mini-ILA Waveform</h1>" + wave_data
    IPython.core.display.display_html(IPython.core.display.HTML(htmldata))

    jsdata = "WaveDrom.ProcessAll();"
    IPython.core.display.display_javascript(
        IPython.core.display.Javascript(
            data=jsdata,
            # important to have the slashes the unix way.
            lib=[
                "wavedrom/wavedrom.min.js",
                "wavedrom/skins/default.js",
                "wavedrom/skins/narrow.js",
            ],
        )
    )


def find_mila_cores(cs_url: str, hw_url: str = "", silent: bool = False) -> MILACoreClient:

    if hw_url != "":
        server = connect_xicom(cs_url)
        server.connect_remote(hw_url)
        if not silent:
            print(f"Connected to cs_server {cs_url} and hw_server {hw_url}")
    else:
        # single argument passed in is intended to be the actual cs_server
        server = cs_url

    if not silent:
        print(f"CS Server services: {server.services}")

    # Set up core detection and get list of found cores along with context properties
    cs_view = get_cs_view(server)
    if not silent:
        print(f"Searching for DPC in {cs_view}")
    dpc = None
    for node in cs_view.get_children():
        if "DPC" in node.props.get("Name") or "XVC" in node.ctx:
            dpc = cs_view.get_node(node.ctx, CoreParent)
            break

    # Set up all available hardened cores
    dpc.setup_cores()

    if not silent:
        print()
        print(f"DPC: {dpc}")

        print()
        print("ChipScope View:")
        cs_view.print_tree(False)

    # Find MILAs under DPC, returning as generator
    milas = cs_view.find_nodes(dpc, MILACoreClient)

    return milas


def get_mila_by_index(cs_url: str, hw_url: str, idx: int) -> MILACoreClient:

    milas = find_mila_cores(cs_url, hw_url, True)
    mila_list = list(milas)
    mila_count = len(mila_list)

    if (idx < 0) or not (idx < mila_count):
        raise Exception(
            f"Invalid index given to get target MILA: {idx} \
                          \nTotal number of MILA found is: {mila_count}"
        )

    mila = mila_list[idx]

    return mila



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing
from typing import Any, Dict
from chipscopy.tcf import services
from chipscopy.proxies.CorePropertyProxy import CorePropertyProxy
from chipscopy.utils.logger import log

# NOTE: There should NOT be any dependency from proxy to server.
#  Instead define the service name locally in this file.
# from chipscope.server.harden.ibert.ibert_service import IBERTService

# NOTE: This name MUST match with the service name defined in the server side IBERT code
IBERT_SERVICE_NAME = "IBERT"

# NOTE: PLEASE USE THIS DOMAIN NAME IN ALL LOG MESSAGES FROM THIS FILE.
DOMAIN_NAME = "proxy_ibert"


class IBERTService(services.Service):
    def getName(self) -> str:
        return IBERT_SERVICE_NAME


class IBERTProxy(IBERTService):
    def __init__(self, channel):
        super(IBERTProxy, self).__init__(channel)
        self.listeners = {}

    def initialize(self, node_id: str, done) -> None:
        log[DOMAIN_NAME].debug("Sending initializeCmd")
        return self.send_xicom_command("initialize", (node_id,), done)

    def setup(self, options: Dict[str, Any], done):
        return self.send_xicom_command("setup", (options["node_id"], options), done)

    def get_layout(self, options: Dict[str, Any], done):
        return self.send_xicom_command("get_layout", (options["node_id"],), done)

    def read(self, options: Dict[str, Any], done):
        return self.send_xicom_command("read", (options["node_id"], options), done)

    def write(self, options: Dict[str, Any], done):
        return self.send_xicom_command("write", (options["node_id"], options), done)

    def get_property(self, options: Dict[str, Any], done):
        return self.send_xicom_command("get_property", (options["node_id"], options), done)

    def get_property_group(self, options: Dict[str, Any], done):
        return self.send_xicom_command("get_property_group", (options["node_id"], options), done)

    def set_property(self, options: Dict[str, Any], done):
        return self.send_xicom_command("set_property", (options["node_id"], options), done)

    def refresh_property(self, options: Dict[str, Any], done):
        return self.send_xicom_command("refresh_property", (options["node_id"], options), done)

    def refresh_property_group(self, options: Dict[str, Any], done):
        return self.send_xicom_command(
            "refresh_property_group", (options["node_id"], options), done
        )

    def commit_property(self, options: Dict[str, Any], done):
        return self.send_xicom_command("commit_property", (options["node_id"], options), done)

    def list_property_groups(self, options: Dict[str, Any], done):
        return self.send_xicom_command("list_property_groups", (options["node_id"],), done)

    def report_property(self, options: Dict[str, Any], done):
        return self.send_xicom_command("report_property", (options["node_id"], options), done)

    def add_to_property_watchlist(self, options: Dict[str, Any], done):
        return self.send_xicom_command(
            "add_to_property_watchlist", (options["node_id"], options), done
        )

    def remove_from_property_watchlist(self, options: Dict[str, Any], done):
        return self.send_xicom_command(
            "remove_from_property_watchlist", (options["node_id"], options), done
        )

    def start_eye_scan(self, options: Dict[str, Any], done):
        return self.send_xicom_command("start_eye_scan", (options["node_id"], options), done)

    def terminate_eye_scan(self, options: Dict[str, Any], done):
        return self.send_xicom_command("terminate_eye_scan", (options["node_id"], options), done)

    def get_eye_scan_parameters(self, options: Dict[str, Any], done):
        return self.send_xicom_command(
            "get_eye_scan_parameters", (options["node_id"], options), done
        )

    def start_yk_scan(self, options: Dict[str, Any], done):
        return self.send_xicom_command("start_yk_scan", (options["node_id"], options), done)

    def terminate_yk_scan(self, options: Dict[str, Any], done):
        return self.send_xicom_command("terminate_yk_scan", (options["node_id"], options), done)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
from enum import Enum, EnumMeta
from io import TextIOBase
from pprint import pformat
from collections import defaultdict
from dataclasses import dataclass, asdict
from re import compile as re_compile
from typing import Union, Dict, List, Optional, Tuple, NewType

from chipscopy.api import CoreType
from chipscopy.utils import Enum2StrEncoder


@dataclass(frozen=True)
class LtxProbe:
    """Represents a single probe in the LTX file"""

    name: str
    direction: str
    probe_type: str
    port_index: int
    port_msb_index: int
    port_lsb_index: int
    is_bus: bool
    bus_left_index: int
    bus_right_index: int
    enum_def: Optional[EnumMeta]


@dataclass(frozen=True)
class LtxCore:
    core_type: CoreType
    cell_name: str
    uuid: str
    debug_hub_address: int
    probes: [LtxProbe]
    upstream_cell_names: [str]
    """ E.g. A VIO debug core would list the connected debug_hub, as an upstream core."""

    def __str__(self) -> str:
        return pformat(self.__dict__)


@dataclass(frozen=True)
class LtxStreamRef:
    """
    Reference to a debug core stream interface.
    A debug core has one or more control/data streams which connect to other debug cores.
    E.g. A debug hub may have multiple downstream connections, one for each connected VIO and ILA debug cores.
    """

    cell_name: str
    uuid: int
    stream_index: int


ParentEnums = NewType("ParentEnums", Dict[int, Dict[str, str]])


class Ltx:
    __LTX_IP_NAME_TO_CORE_TYPE = {
        "axi_dbg_hub": CoreType.AXI_DEBUG_HUB,
        "axis_ila": CoreType.AXIS_ILA,
        "axis_vio": CoreType.AXIS_VIO,
        "trace": CoreType.AXIS_TRACE,
    }

    __NULL_UUID = "0" * 32
    __net_bus_index_re = None
    __subnet_bus_index_re = None

    def __init__(self):
        self._debug_hub_addresses: [int] = []
        self._cores: {CoreType, [LtxCore]} = defaultdict(list)
        # key: upstream_cell_name. value: list of downstream LtxStreamRef
        # E.g. an entry for a debug_hub core, would list other debug cores connected to it.
        self._downstream_refs: {str: List[LtxStreamRef]} = defaultdict(list)
        self._parent_id_to_enum: ParentEnums = {}

    def get_debug_hub_addresses(self) -> [int]:
        return self._debug_hub_addresses

    def get_core(self, core_type: CoreType, uuid: str = "") -> Optional[LtxCore]:
        """ Return first matching core, or None."""
        cores = [core for core in self._cores[core_type] if core.uuid.upper() == uuid.upper()]
        return cores[0] if cores else None

    def get_downstream_refs(self, upstream_core_name: str) -> List[LtxStreamRef]:
        """
        Returns list of LtxStreamRef, which lists which debug cores the upstream core is connected to.
        The list is not ordered.
        E.g. given a debug_hub cell_name, the function returns a list of connected debug cores.
        """
        return self._downstream_refs.get(upstream_core_name, [])

    def get_cores(self, core_type: CoreType) -> [LtxCore]:
        return self._cores[core_type]

    def parse_file(self, probes_file: str) -> None:
        with open(probes_file, "r") as json_file:
            self.parse_ltx(json_file)

    def parse_ltx(self, fh: TextIOBase) -> None:
        json_data = json.load(fh)

        ltx_root = json_data.get("ltx_root", None)
        ltx_data = ltx_root.get("ltx_data", []) if ltx_root else []
        # Always expect one item in 'ltx_data' list.
        ltx_data_item = ltx_data[0] if ltx_data else None
        debug_cores = ltx_data_item.get("debug_cores", None) if ltx_data_item else None
        if not debug_cores:
            return
        if "parent_nets" in ltx_data_item:
            self._parent_id_to_enum = self._read_parent_net_enums(ltx_data_item["parent_nets"])
        for debug_core in debug_cores:
            core = Ltx._read_core(debug_core, self._parent_id_to_enum)
            if not core:
                continue

            self._cores[core.core_type].append(core)
            for idx, up_stream_cell_names in enumerate(core.upstream_cell_names):
                self._downstream_refs[up_stream_cell_names].append(
                    LtxStreamRef(core.cell_name, core.uuid, idx)
                )

        self._debug_hub_addresses = [
            int(core.debug_hub_address, 16) for core in self._cores[CoreType.AXI_DEBUG_HUB]
        ]

    def verify_port_width_by_type(
        self, core_type: CoreType, uuid: str, width_by_type: Dict[str, List[int]]
    ) -> None:
        mismatch_msg = "Mismatch between LTX file(s) and device debug core.\n"
        core = self.get_core(core_type, uuid)
        if not core:
            raise KeyError(
                mismatch_msg
                + f"Device {core_type.value} core with UUID {uuid} is not found in the LTX file."
            )

        for probe in core.probes:
            port_widths = width_by_type.get(probe.probe_type, [])
            port_width = port_widths[probe.port_index] if len(port_widths) > probe.port_index else 0
            if port_width == 0:
                raise KeyError(
                    mismatch_msg
                    + f'LTX core "{core.cell_name}" probe "{probe.name}" of type {probe.probe_type} '
                    f'refers to core port index "{probe.port_index}", '
                    f"which does not exist in the debug core for probe type {probe.probe_type}."
                )
            if port_width <= probe.port_msb_index:
                raise ValueError(
                    mismatch_msg
                    + f'LTX core "{core.cell_name}" probe "{probe.name}" of type {probe.probe_type} '
                    f"has invalid msb index:{probe.port_msb_index}.\n The port width is {port_width}"
                )

    @staticmethod
    def _read_parent_net_enums(parent_nets: []) -> ParentEnums:
        """State values are returned as binary strings, so the bit_width is obvious.
        Return type: {parent_id: {state_name, state_binary_str_value}}
        """
        pid_to_enums = {}
        for pnet in parent_nets:
            pid_states = {}
            for sr in pnet.get("state_reg", []):
                pid_states.update((state["name"], state["val"]) for state in sr.get("states", []))
            if pid_states:
                pid_to_enums[pnet["id"]] = pid_states
        return pid_to_enums

    @staticmethod
    def _get_enum_def(
        probe_name: str, parent_ids: [int], parent_enums: ParentEnums, probe_width: int
    ) -> Optional[Enum]:
        if not parent_ids:
            return None
        p_enum = parent_enums.get(parent_ids[0], None)
        if not all([len(val) == probe_width for val in p_enum.values()]):
            # Old LTX files have bad enums, for non-important probes. Skip enum.
            return None
        enum_vals = {state: int(val, 2) for state, val in p_enum.items()}
        return Enum(probe_name, enum_vals)

    @staticmethod
    def _read_core(in_core: {}, parent_enums: ParentEnums) -> Union[LtxCore, None]:
        pins = in_core.get("pins", [])
        probes = [Ltx._read_pin(pin, parent_enums) for pin in pins]
        probes = [probe for probe in probes if probe]
        ip_name = in_core.get("ipName", "")
        if ip_name == "axi_dbg_hub" and "address_info" in in_core:
            addr = in_core["address_info"].get("offset", "")
        else:
            addr = ""

        core_type = Ltx.__LTX_IP_NAME_TO_CORE_TYPE.get(ip_name, "")
        cell_name = in_core.get("name", "")
        uuid = in_core.get("uuid", Ltx.__NULL_UUID).upper()
        upstream = in_core.get("master", "")
        upstream_cell_names = upstream.split(":") if upstream else []

        # todo syntax/semantic checks?
        if not core_type:
            return None
        return LtxCore(core_type, cell_name, uuid, addr, probes, upstream_cell_names)

    @staticmethod
    def read_nets(nets: []) -> Tuple[bool, int, int, List[int]]:
        """
        Returns (name, left index, right index, [parent_ids])
        """
        if not nets:
            return False, 0, 0, []

        parent_ids = [net.get("parentNetId", None) for net in nets]
        parent_ids = [p_id for p_id in parent_ids if p_id is not None]

        left_name = nets[0].get("name", "")
        right_name = nets[-1].get("name", "")

        if not Ltx.__net_bus_index_re:
            Ltx.__net_bus_index_re = re_compile(r"(.+)\[(-?\d+)]$")
        left_m = Ltx.__net_bus_index_re.match(left_name)
        right_m = Ltx.__net_bus_index_re.match(right_name)
        if not left_m or not right_m:
            return False, 0, 0, parent_ids
        return left_m.group(1), int(left_m.group(2)), int(right_m.group(2)), parent_ids

    @staticmethod
    def read_subnets(subnets: []) -> Tuple[int, int, List[int]]:
        """
        If a bus returns (left index, right index, [parent_ids])
        """
        if not subnets:
            return 0, 0, []

        parent_ids = [subnet.get("parentNetId", None) for subnet in subnets]
        parent_ids = [p_id for p_id in parent_ids if p_id is not None]

        left_name = subnets[0].get("name", "")
        right_name = subnets[-1].get("name", "")

        if not Ltx.__subnet_bus_index_re:
            Ltx.__subnet_bus_index_re = re_compile(r".+\[(-?\d+)]$")
        left_m = Ltx.__subnet_bus_index_re.match(left_name)
        right_m = Ltx.__subnet_bus_index_re.match(right_name)
        if not left_m or not right_m:
            return 0, 0, parent_ids
        return int(left_m.group(1)), int(right_m.group(1)), parent_ids

    @staticmethod
    def _read_pin(pin: {}, parent_enums: ParentEnums) -> LtxProbe:
        nets = pin.get("nets", [])
        if len(nets) == 1:
            name = nets[0].get("name", "") if nets else ""
            subnets = nets[0].get("subnets", [])
            bus_left_index, bus_right_index, parent_ids = Ltx.read_subnets(subnets)
        else:
            # Some state machine registers are split up in multiple nets.
            name, bus_left_index, bus_right_index, parent_ids = Ltx.read_nets(nets)

        probe_type = pin.get("type", "")
        is_bus = pin.get("isVector", False)
        port_index = pin.get("portIndex", 0)
        # The LTX file has port right/left switched.
        port_msb_index = pin.get("rightIndex", 0)
        port_lsb_index = pin.get("leftIndex", 0)
        direction = pin.get("direction", "")
        probe_def = Ltx._get_enum_def(
            name, parent_ids, parent_enums, abs(bus_left_index - bus_right_index) + 1
        )
        return LtxProbe(
            name,
            direction,
            probe_type,
            port_index,
            port_msb_index,
            port_lsb_index,
            is_bus,
            bus_left_index,
            bus_right_index,
            probe_def,
        )

    def __str__(self) -> str:
        dump = "\ndebug_hub_addresses :" + ",".join(
            [hex(addr) for addr in self._debug_hub_addresses]
        )
        for core_type, cores in self._cores.items():
            underline = "=" * (len(core_type.__str__()) + 4)
            for idx, core in enumerate(cores):
                dump += f"\n\n{core_type}:{idx:3}\n{underline}\n"
                dump += core.__str__() + "\n"

        return dump

    def __repr__(self) -> str:
        cores = defaultdict(list)
        for core_type, core_list in self._cores.items():
            for core in core_list:
                cores[str(core_type)].append(asdict(core))
        downstream_refs = defaultdict(list)
        for up, down_list in self._downstream_refs.items():
            for down in down_list:
                downstream_refs[up].append(asdict(down))

        ret_dict = {
            "debug_hub_addresses": [f"0x{addr:x}" for addr in self._debug_hub_addresses],
            "downstream_refs": downstream_refs,
            "parent_id_to_enum": self._parent_id_to_enum,
            "cores": cores,
        }
        json_dict = json.dumps(ret_dict, cls=Enum2StrEncoder, indent=4)
        return json_dict



# *****************************************************************************
# * Copyright (c) 2020 Xilinx, Inc.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Xilinx
# *****************************************************************************

import re
import base64

DEFAULT_ARG_VERSION = "arg_3"

data_count_re = re.compile('\((\d+)\)')


class JsonDataEncoder(object):
    def __init__(self, data):
        self.data = data

    def __json__(self):
        return "({})".format(len(self.data))

    def __str__(self):
        return self.__json__()

    def __repr__(self):
        return self.__json__()

def preprocess_list(items):
    l = []
    data = bytearray()
    for v in items:
        t = type(v)
        if t == dict:
            v, d = preprocess_props(v)
            data += d
        elif t == bytearray or t == bytes:
            data += v
            v = JsonDataEncoder(v)
        l.append(v)
    return l, data


def preprocess_props(props):
    out_props = {}
    data = bytearray()
    for k, v in props.items():
        t = type(v)
        if t == bytearray or t == bytes:
            data += v
            v = JsonDataEncoder(v)
        elif t == dict:
            v, d = preprocess_props(v)
            data += d
        elif t == list:
            v, d = preprocess_list(v)
            data += d
        out_props[k] = v
    return out_props, data


def args_from_props(**props):
    args, data = preprocess_props(props)
    return DEFAULT_ARG_VERSION, args, data


def to_xargs(args):
    xargs = []
    if type(args) == dict:
        props, data = preprocess_props(args)
        xargs.append(props)
        if len(data) > 0:
            xargs.append(data)
    else:
        for arg in args:
            if type(arg) == dict:
                props, data = preprocess_props(arg)
                xargs.append(props)
                if len(data) > 0:
                    xargs.append(data)
            elif isinstance(arg, list) or isinstance(arg, tuple):
                arg, data = preprocess_list(arg)
                xargs.append(arg)
                if len(data) > 0:
                    xargs.append(data)
            else:
                xargs.append(arg)
    return xargs


class ArgDataParser(object):
    def __init__(self, data):
        self.data = data
        self.pos = 0

    def __call__(self, size):
        end = self.pos + size
        if not isinstance(self.data, bytearray) and not isinstance(self.data, bytes):
            self.data = base64.b64decode(self.data)
        chunk = self.data[self.pos: end]
        self.pos = end
        return chunk


def _dict_from_args(args, parser):
    return {k: _handle_arg_value(v, parser) for k, v in args.items()}


def _handle_arg_value(arg_val, parser):
    t = type(arg_val)
    if t == dict:
        arg_val = _dict_from_args(arg_val, parser)
    elif t == list:
        arg_val = [_handle_arg_value(i, parser) for i in arg_val]
    elif t == str and parser:
        m = data_count_re.match(arg_val)
        if m:
            arg_val = parser(int(m.groups()[0]))
    return arg_val


def props_from_args(args):
    props = from_xargs(args)

    if props and len(props) == 1:
        props = props[0]

    return props


def from_xargs(args):
    props = []
    parser = None

    for i in range(len(args)):
        if args[i] == DEFAULT_ARG_VERSION:
            continue
        elif parser:
            parser = None
            continue
        elif i + 1 < len(args) and \
                (type(args[i + 1]) == bytearray or type(args[i+1]) == str):
            parser = ArgDataParser(args[i + 1])
            props.append(_handle_arg_value(args[i], parser))
            if parser.pos == 0:  # parser unused so don't skip
                parser = None
        # elif type(args[i]) == dict:
        #     if i+1 < len(args) and type(args[i+1]) == bytearray:
        #         parser = ArgDataParser(args[i+1])
        #     props.append(_dict_from_args(args[i], parser))
        else:
            props.append(args[i])
    return props



# *****************************************************************************
# * Copyright (c) 2020 Xilinx, Inc.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Xilinx
# *****************************************************************************

from typing import List, Dict, Any
from chipscopy.tcf.services import Service, DoneHWCommand

NAME = "Test"
"""Test service name."""


class TestListener(object):
    """Test event listener is notified when test service needs to send an internal event
    """

    def exiting(self):
        """Called when an exit command is received.  Used to gracefully shuttdown during testing when a client is done.
        """
        pass


class TestService(Service):
    """TCF Test service interface."""
    def getName(self):
        """Get this service name.

        :returns: The value of string :const:`NAME`
        """
        return NAME

    def add_manager(self, manager_name: str, done: DoneHWCommand):
        """
        Adds manager of a given name to server
        :param manager_name: Name to add
        :param done: Callback when command is complete
        """
        raise NotImplementedError("Abstract method")

    def remove_manager(self, manager_name: str, done: DoneHWCommand):
        """
        Removes manager of a given name from server
        :param manager_name: Name to remove
        :param done: Callback when command is complete
        """
        raise NotImplementedError("Abstract method")

    def add_node(self, manager_name: str, ctx: str, props: Dict, done: DoneHWCommand):
        """
        Adds node to manager with given props.  If property, ParentID, is set the node will be added as a child of
        node with that context id.

        :param manager_name: Name of manager to which to add the node
        :param ctx: Node context handle to use
        :param props: Properties to set in the node
        :param done: Callback when command is complete
        """
        raise NotImplementedError("Abstract method")

    def remove_node(self, manager_name: str, ctx: str, done: DoneHWCommand):
        """
        Removes node from manager.

        :param manager_name: Name of manager from which to remove
        :param ctx: Node context handle to use
        :param done: Callback when command is complete
        """
        raise NotImplementedError("Abstract method")

    def update_node(self, manager_name: str, ctx: str, props: Dict, done: DoneHWCommand):
        """
        Updates node with given props.

        :param manager_name: Name of manager
        :param ctx: Node context handle to use
        :param props: Properties to set in the node
        :param done: Callback when command is complete
        """
        raise NotImplementedError("Abstract method")

    def exit(self, done: DoneHWCommand):
        """
        Tells the server to exit.

        :param done: Callback when command is complete
        """
        raise NotImplementedError("Abstract method")

